from lenstronomy.GalKin.observation import GalkinObservation
from lenstronomy.GalKin.galkin_model import GalkinModel

import numpy as np

__all__ = ["Galkin"]


class Galkin(GalkinModel, GalkinObservation):
    """Major class to compute velocity dispersion measurements given light and mass
    models.

    The class supports any mass and light distribution (and superposition thereof) that has a 3d correspondance in their
    2d lens model distribution. For models that do not have this correspondance, you may want to apply a
    Multi-Gaussian Expansion (MGE) on their models and use the MGE to be de-projected to 3d.

    The computation follows Mamon&Lokas 2005 and performs the spectral rendering of the seeing convolved apperture with
    the method introduced by Birrer et al. 2016.

    The class supports various types of anisotropy models (see Anisotropy class) and aperture types
    (see Aperture class).

    Solving the Jeans Equation requires a numerical integral over the 3d light and mass profile (see Mamon&Lokas 2005).
    This class (as well as the dedicated LightModel and MassModel classes) perform those integral numerically with an
    interpolated grid.

    The seeing convolved integral over the aperture is computed by rendering spectra (light weighted LOS kinematics)
    from the light distribution.

    The cosmology assumed to compute the physical mass and distances are set via the kwargs_cosmo keyword arguments.
        d_d: Angular diameter distance to the deflector (in Mpc)
        d_s: Angular diameter distance to the source (in Mpc)
        d_ds: Angular diameter distance from the deflector to the source (in Mpc)

    The numerical options can be chosen through the kwargs_numerics keywords

        interpol_grid_num: number of interpolation points in the light and mass profile (radially). This number should
        be chosen high enough to accurately describe the true light profile underneath.
        log_integration: bool, if True, performs the interpolation and numerical integration in log-scale.

        max_integrate: maximum 3d radius to where the numerical integration of the Jeans Equation solver is made.
        This value should be large enough to contain most of the light and to lead to a converged result.
        min_integrate: minimal integration value. This value should be very close to zero but some mass and light
        profiles are diverging and a numerically stabel value should be chosen.

    These numerical options should be chosen to allow for a converged result (within your tolerance) but not too
    conservative to impact too much the computational cost. Reasonable values might depend on the specific problem.
    """

    def __init__(
        self,
        kwargs_model,
        kwargs_aperture,
        kwargs_psf,
        kwargs_cosmo,
        kwargs_numerics=None,
        analytic_kinematics=False,
    ):
        """

        :param kwargs_model: keyword arguments describing the model components
        :param kwargs_aperture: keyword arguments describing the spectroscopic aperture, see Aperture() class
        :param kwargs_psf: keyword argument specifying the PSF of the observation
        :param kwargs_cosmo: keyword arguments that define the cosmology in terms of the angular diameter distances
         involved
        :param kwargs_numerics: numerics keyword arguments
        :param analytic_kinematics: bool, if True uses the analytic kinematic model
        """
        GalkinModel.__init__(
            self,
            kwargs_model,
            kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
            analytic_kinematics=analytic_kinematics,
        )
        GalkinObservation.__init__(
            self, kwargs_aperture=kwargs_aperture, kwargs_psf=kwargs_psf
        )

    def dispersion(
        self, kwargs_mass, kwargs_light, kwargs_anisotropy, sampling_number=1000
    ):
        """Computes the averaged LOS velocity dispersion in the slit (convolved)

        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :param sampling_number: int, number of spectral sampling of the light
            distribution
        :return: integrated LOS velocity dispersion in units [km/s]
        """
        sigma2_IR_sum = 0
        IR_sum = 0
        for i in range(0, sampling_number):
            sigma2_IR, IR = self._draw_one_sigma2(
                kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            sigma2_IR_sum += sigma2_IR
            IR_sum += IR
        sigma_s2_average = sigma2_IR_sum / IR_sum
        # apply unit conversion from arc seconds and deflections to physical velocity dispersion in (km/s)
        self.numerics.delete_cache()
        return np.sqrt(sigma_s2_average) / 1000.0  # in units of km/s

    def dispersion_map(
        self,
        kwargs_mass,
        kwargs_light,
        kwargs_anisotropy,
        num_kin_sampling=1000,
        num_psf_sampling=100,
    ):
        """Computes the velocity dispersion in each Integral Field Unit.

        :param kwargs_mass: keyword arguments of the mass model
        :param kwargs_light: keyword argument of the light model
        :param kwargs_anisotropy: anisotropy keyword arguments
        :param num_kin_sampling: int, number of draws from a kinematic prediction of a
            LOS
        :param num_psf_sampling: int, number of displacements/render from a spectra to
            be displaced on the IFU
        :return: ordered array of velocity dispersions [km/s] for each unit
        """
        # draw from light profile (3d and 2d option)
        # compute kinematics of it (analytic or numerical)
        # displace it n-times
        # add it and keep track of how many draws are added on each segment
        # compute average in each segment
        # return value per segment
        num_segments = self.num_segments
        sigma2_IR_sum = np.zeros(num_segments)
        count_draws = np.zeros(num_segments)

        for i in range(0, num_kin_sampling):
            r, R, x, y = self.numerics.draw_light(kwargs_light)
            sigma2_IR, IR = self.numerics.sigma_s2(
                r, R, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            for k in range(0, num_psf_sampling):
                x_, y_ = self.displace_psf(x, y)
                bool_ap, ifu_index = self.aperture_select(x_, y_)
                if bool_ap is True:
                    sigma2_IR_sum[ifu_index] += sigma2_IR
                    count_draws[ifu_index] += IR

        sigma_s2_average = sigma2_IR_sum / count_draws
        # apply unit conversion from arc seconds and deflections to physical velocity dispersion in (km/s)
        self.numerics.delete_cache()
        return np.sqrt(sigma_s2_average) / 1000.0  # in units of km/s

    def _draw_one_sigma2(self, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """

        :param kwargs_mass: mass model parameters (following lenstronomy lens model conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to anisotropy type chosen.
            We refer to the Anisotropy() class for details on the parameters.
        :return: integrated LOS velocity dispersion in angular units for a single draw of the light distribution that
         falls in the aperture after displacing with the seeing
        """
        while True:
            r, R, x, y = self.numerics.draw_light(kwargs_light)
            x_, y_ = self.displace_psf(x, y)
            bool_ap, _ = self.aperture_select(x_, y_)
            if bool_ap is True:
                break
        sigma2_IR, IR = self.numerics.sigma_s2(
            r, R, kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        return sigma2_IR, IR



# %%
"""Provisional Roman instrument and observational settings."""
import lenstronomy.Util.util as util
import os
import astropy.io.fits as pyfits

# magnitude_zero_point: table 1 from https://iopscience.iop.org/article/10.3847/1538-4357/aac08b/pdf
# ccd_gain found right under table 1 in paper
# seeing, read_noise, pixel_scale (labelled as plate_scale on website): https://roman.gsfc.nasa.gov/science/WFI_technical.html
# sky brightness calculated using count rates per pixel at minimum Zodiacal light given in website above
# For microlensing survey mode: exposure time, number of exposures for F146 and F087: table 1 with the mission design of WFIRST Cycle 7 from https://iopscience.iop.org/article/10.3847/1538-4365/aafb69/meta#apjsaafb69t1fnd
# For wide area survey mode: exposure time and number of exposures for relevant filters set as given in https://roman.gsfc.nasa.gov/high_latitude_wide_area_survey.html


__all__ = ["Roman"]

F062_band_obs = {
    "sky_brightness": 23.19,
    "magnitude_zero_point": 26.56,
    "seeing": 0.058,
}

F087_band_obs = {
    "sky_brightness": 22.93,
    "magnitude_zero_point": 26.30,
    "seeing": 0.073,
}

F106_band_obs = {
    "sky_brightness": 22.99,
    "magnitude_zero_point": 26.44,
    "seeing": 0.087,
}

F129_band_obs = {
    "sky_brightness": 22.99,
    "magnitude_zero_point": 26.40,
    "seeing": 0.105,
}

F158_band_obs = {
    "sky_brightness": 23.10,
    "magnitude_zero_point": 26.43,
    "seeing": 0.127,
}

F184_band_obs = {
    "sky_brightness": 23.22,
    "magnitude_zero_point": 25.95,
    "seeing": 0.151,
}

F146_band_obs = {
    "sky_brightness": 22.03,
    "magnitude_zero_point": 26.65,
    "seeing": 0.105,
}

# - keyword sky_brightness: sky brightness (in magnitude per square arcseconds in units of electrons)
# - keyword magnitude_zero_point: magnitude in which 1 count (e-) per second per arcsecond square is registered
# - keyword seeing: Full-Width-at-Half-Maximum (FWHM) of PSF
# - keyword psf_type: string, type of PSF ('GAUSSIAN' supported)


class Roman(object):
    """Class contains Roman instrument and observation configurations."""

    def __init__(self, band="F062", psf_type="GAUSSIAN", survey_mode="wide_area"):
        """

        :param band: string, 'F062', 'F087', 'F106', 'F129', 'F158' , 'F184' or 'F146' supported. Determines obs dictionary.
        :param psf_type: string, type of PSF ('GAUSSIAN', 'PIXEL' supported).
        """

        if band == "F062":
            self.obs = F062_band_obs
        elif band == "F087":
            self.obs = F087_band_obs
        elif band == "F106":
            self.obs = F106_band_obs
        elif band == "F129":
            self.obs = F129_band_obs
        elif band == "F158":
            self.obs = F158_band_obs
        elif band == "F184":
            self.obs = F184_band_obs
        elif band == "F146":
            self.obs = F146_band_obs
        else:
            raise ValueError(
                "band %s not supported! Choose 'F062', 'F087', 'F106', 'F129', 'F158' , 'F184' or 'F146'"
                % band
            )

        if survey_mode == "wide_area":
            # the number of exposures is given per sector
            # a full pass of the High Latitude Wide Area Survey is 155 sectors

            if band in ["F106", "F158", "F184", "F062"]:
                exp_per_tile = 3
            elif band == "F129":
                exp_per_tile = 4
            else:
                raise ValueError(
                    "band %s is not supported with the wide_area survey mode! Choose 'F106', 'F062, 'F158', 'F184' or F129"
                    % band
                )

            self.obs.update({"exposure_time": 146, "num_exposures": 32 * exp_per_tile})
        elif survey_mode == "microlensing":
            if band == "F146":
                # These are the exposure times and number of exposures for the primary filter, F146
                self.obs.update({"exposure_time": 46.8, "num_exposures": 41000})
            elif band == "F087":
                # These are the exposure times and number of exposures for the secondary filter, F087
                self.obs.update({"exposure_time": 286.0, "num_exposures": 860})
            else:
                raise ValueError(
                    "band %s is not supported with the microlensing survey mode! Choose 'F146' or 'F087'"
                    % band
                )
        else:
            raise ValueError(
                "survey mode %s not supported! Choose 'wide_area' or 'microlensing'"
                % survey_mode
            )

        if psf_type == "PIXEL":
            import lenstronomy

            module_path = os.path.dirname(lenstronomy.__file__)
            psf_filename = os.path.join(
                module_path,
                "SimulationAPI/ObservationConfig/PSF_models/{}.fits".format(band),
            )
            kernel = pyfits.getdata(psf_filename)
            self.obs.update({"psf_type": "PIXEL", "kernel_point_source": kernel})
        elif psf_type == "GAUSSIAN":
            self.obs.update({"psf_type": "GAUSSIAN"})
        else:
            raise ValueError("psf_type %s not supported!" % psf_type)

        self.camera = {"read_noise": 15.5, "pixel_scale": 0.11, "ccd_gain": 1}

        # - keyword read_noise: std of noise generated by read-out (in units of electrons)
        # - keyword pixel_scale: scale (in arcseconds) of pixels
        # - keyword ccd_gain: electrons/ADU (analog-to-digital unit)

    def kwargs_single_band(self):
        """

        :return: merged kwargs from camera and obs dicts
        """
        kwargs = util.merge_dicts(self.camera, self.obs)
        return kwargs



from lenstronomy.Util.magnification_finite_util import (
    auto_raytracing_grid_size,
    auto_raytracing_grid_resolution,
    setup_mag_finite,
)
from lenstronomy.LensModel.lens_model import LensModel
import numpy.testing as npt
import pytest


class TestMagnificationFiniteUtil(object):
    def test_grid_resolution(self):
        source_size = 1.0
        res_scale = 0.03
        ref = 15.0
        power = 1.2
        resolution = auto_raytracing_grid_resolution(source_size, res_scale, ref, power)

        npt.assert_almost_equal(resolution, res_scale * (source_size / ref) ** power)

    def test_grid_size(self):
        source_size = 20.0
        size = auto_raytracing_grid_size(source_size, grid_size_scale=0.04, power=0.9)

        npt.assert_almost_equal(size, 0.04 * source_size**0.9)

    def test_setup(self):
        cosmo = None
        lens_model = LensModel(["EPL"])
        grid_radius_arcsec = None
        grid_resolution = None
        source_fwhm_parsec = 30.0
        source_light_model = "SINGLE_GAUSSIAN"
        z_source = 2.0
        source_x, source_y = 0.0, 0.0
        dx, dy, amp_scale, size_scale = None, None, None, None
        (
            gridx,
            gridy,
            source_model,
            kwargs_source,
            grid_resolution,
            grid_radius_arcsec,
        ) = setup_mag_finite(
            cosmo,
            lens_model,
            grid_radius_arcsec,
            grid_resolution,
            source_fwhm_parsec,
            source_light_model,
            z_source,
            source_x,
            source_y,
            dx,
            dy,
            amp_scale,
            size_scale,
        )
        npt.assert_equal(True, len(source_model.func_list) == 1)
        npt.assert_equal(True, grid_resolution is not None)
        npt.assert_equal(True, grid_radius_arcsec is not None)

        grid_resolution = 0.001
        grid_radius_arcsec = 0.05
        dx, dy, amp_scale, size_scale = 0.0, 0.1, 1.0, 1.0
        source_light_model = "DOUBLE_GAUSSIAN"
        (
            gridx,
            gridy,
            source_model,
            kwargs_source,
            grid_resolution,
            grid_radius_arcsec,
        ) = setup_mag_finite(
            cosmo,
            lens_model,
            grid_radius_arcsec,
            grid_resolution,
            source_fwhm_parsec,
            source_light_model,
            z_source,
            source_x,
            source_y,
            dx,
            dy,
            amp_scale,
            size_scale,
        )
        npt.assert_equal(True, len(source_model.func_list) == 2)
        npt.assert_equal(
            kwargs_source[1]["center_y"], kwargs_source[0]["center_y"] + dy
        )
        npt.assert_equal(
            kwargs_source[1]["center_x"], kwargs_source[0]["center_x"] + dx
        )
        npt.assert_equal(True, grid_resolution == 0.001)
        npt.assert_equal(True, grid_radius_arcsec == 0.05)

        source_light_model = "trash"
        npt.assert_raises(
            Exception,
            setup_mag_finite,
            cosmo,
            lens_model,
            grid_radius_arcsec,
            grid_resolution,
            source_fwhm_parsec,
            source_light_model,
            z_source,
            source_x,
            source_y,
            dx,
            dy,
            amp_scale,
            size_scale,
        )


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import copy
from lenstronomy.Util import class_creator
from lenstronomy.ImSim.image2source_mapping import Image2SourceMapping
from lenstronomy.LensModel.Solver.solver import Solver
from lenstronomy.LensModel.lens_param import LensParam
from lenstronomy.LightModel.light_param import LightParam
from lenstronomy.PointSource.point_source_param import PointSourceParam
from lenstronomy.Sampling.special_param import SpecialParam

__all__ = ["Param"]


class Param(object):
    """Class that handles the parameter constraints. In particular when different model
    profiles share joint constraints.

    Options between same model classes:

    'joint_lens_with_lens':list [[i_lens, k_lens, ['param_name1', 'param_name2', ...]], [...], ...],
    joint parameter between two lens models

    'joint_lens_light_with_lens_light':list [[i_lens_light, k_lens_light, ['param_name1', 'param_name2', ...]], [...], ...],
    joint parameter between two lens light models, the second adopts the value of the first

    'joint_source_with_source':list [[i_source, k_source, ['param_name1', 'param_name2', ...]], [...], ...],
    joint parameter between two source surface brightness models, the second adopts the value of the first


    Options between different model classes:

    'joint_lens_with_light': list [[i_light, k_lens, ['param_name1', 'param_name2', ...]], [...], ...],
    joint parameter between lens model and lens light model

    'joint_source_with_point_source': list [[i_point_source, k_source], [...], ...],
    joint position parameter between source light model and point source

    'joint_lens_light_with_point_source': list [[i_point_source, k_lens_light], [...], ...],
    joint position parameter between lens model and lens light model

    'joint_extinction_with_lens_light': list [[i_lens_light, k_extinction, ['param_name1', 'param_name2', ...]], [...], ...],
    joint parameters between the lens surface brightness and the optical depth models

    'joint_lens_with_source_light': [[i_source, k_lens, ['param_name1', 'param_name2', ...]], [...], ...],
    joint parameter between lens model and source light model. Samples light model parameter only.

    'mass_scaling_list': e.g. [False, 1, 1, False, 2, False, 1, ...]
    Links lens models to have their masses scaled together. In this example,
    masses with False are not scaled, masses labeled 1 are scaled together,
    and those labeled 2 are scaled together independent of 1, etc.

    'general_scaling': { 'param1': [False, 1, 1, False, 1, ...], 'param2': [1, 1, 1, False, 2, 2, ...] }
    Generalized parameter scaling. Input should be a dictionary mapping
    parameter names to the masks defining which lens models are scaled together,
    in the same format as for 'mass_scaling_list'. For each scaled parameter,
    two special params will be added called '${param}_scale_factor' and
    '${param}_scale_pow', defining the scaling and power-law of each.

    Each scale will be modified as `param = param_scale_factor * param**param_scale_pow`.

    For example, say we want to jointly constrain the `sigma0` and `Rs` parameters
    of some lens models indexed by `i`, like so:

    .. math::

        \\sigma_{0,i} = \\sigma_0^{ref} L_i^\\alpha \\\\
        r_{cut,i} = r_{cut}^{ref} L_i^\\beta

    To do this we can add the following. The lens models corresponding to
    entries of `1` will be scaled together, and those corresponding to `False`
    will not be. As in `mass_scaling_list`, subsets of models can be scaled
    independently by marking them `2`, `3`, etc.

    >>> 'general_scaling': {
    >>>     'sigma0': [False, 1, 1, False, 1, ...],
    >>>     'Rs': [False, 1, 1, False, 1, ...],
    >>> }

    Then we can choose to fix the power-law and vary the scale factor like so:

    >>> fixed_special = {'sigma0_scale_pow': [alpha*2], 'Rs_scale_pow': [beta]}
    >>> kwargs_special_init = {'sigma0_scale_factor': [17.0], 'Rs_scale_factor': [8]}
    >>> kwargs_special_sigma = {'sigma0_scale_factor': [10.0], 'Rs_scale_factor': [3]}
    >>> kwargs_lower_special = {'sigma0_scale_factor': [0.5], 'Rs_scale_factor': [1]}
    >>> kwargs_upper_special = {'sigma0_scale_factor': [40], 'Rs_scale_factor': [20]}

    >>> special_params = [kwargs_special_init, kwargs_special_sigma, fixed_special, kwargs_lower_special, kwargs_upper_special]

    hierarchy is as follows:
    1. Point source parameters are inferred
    2. Lens light joint parameters are set
    3. Lens model joint constraints are set
    4. Lens model solver is applied
    5. Joint source and point source is applied

    Alternatively to the format of the linking of parameters with IDENTICAL names as listed above as:
    [[i_1, k_2, ['param_name1', 'param_name2', ...]], [...], ...]
    the following format of the arguments are supported to join parameters with DIFFERENT names:
    [[i_1, k_2, {'param_old1': 'param_new1', 'ra_0': 'center_x'}], [...], ...]\

    Log10 sampling of the lens parameters :
    'log_sampling_lens': [[i_lens, ['param_name1', 'param_name2', ...]], [...], ...],
    Sample the log10 of the lens model parameters.
    """

    def __init__(
        self,
        kwargs_model,
        kwargs_fixed_lens=None,
        kwargs_fixed_source=None,
        kwargs_fixed_lens_light=None,
        kwargs_fixed_ps=None,
        kwargs_fixed_special=None,
        kwargs_fixed_extinction=None,
        kwargs_lower_lens=None,
        kwargs_lower_source=None,
        kwargs_lower_lens_light=None,
        kwargs_lower_ps=None,
        kwargs_lower_special=None,
        kwargs_lower_extinction=None,
        kwargs_upper_lens=None,
        kwargs_upper_source=None,
        kwargs_upper_lens_light=None,
        kwargs_upper_ps=None,
        kwargs_upper_special=None,
        kwargs_upper_extinction=None,
        kwargs_lens_init=None,
        linear_solver=True,
        joint_lens_with_lens=[],
        joint_lens_light_with_lens_light=[],
        joint_source_with_source=[],
        joint_lens_with_light=[],
        joint_source_with_point_source=[],
        joint_lens_light_with_point_source=[],
        joint_extinction_with_lens_light=[],
        joint_lens_with_source_light=[],
        mass_scaling_list=None,
        point_source_offset=False,
        general_scaling=None,
        num_point_source_list=None,
        image_plane_source_list=None,
        solver_type="NONE",
        Ddt_sampling=None,
        source_size=False,
        num_tau0=0,
        lens_redshift_sampling_indexes=None,
        source_redshift_sampling_indexes=None,
        source_grid_offset=False,
        kinematic_sampling=None,
        num_shapelet_lens=0,
        log_sampling_lens=[],
    ):
        """

        :param kwargs_model: keyword arguments to describe all model components used in class_creator.create_class_instances()
        :param kwargs_fixed_lens: fixed parameters for lens model (keyword argument list)
        :param kwargs_fixed_source: fixed parameters for source model (keyword argument list)
        :param kwargs_fixed_lens_light: fixed parameters for lens light model (keyword argument list)
        :param kwargs_fixed_ps: fixed parameters for point source model (keyword argument list)
        :param kwargs_fixed_special: fixed parameters for special model parameters (keyword arguments)
        :param kwargs_fixed_extinction: fixed parameters for extinction model parameters (keyword argument list)
        :param kwargs_lower_lens: lower limits for parameters of lens model (keyword argument list)
        :param kwargs_lower_source: lower limits for parameters of source model (keyword argument list)
        :param kwargs_lower_lens_light: lower limits for parameters of lens light model (keyword argument list)
        :param kwargs_lower_ps: lower limits for parameters of point source model (keyword argument list)
        :param kwargs_lower_special: lower limits for parameters of special model parameters (keyword arguments)
        :param kwargs_lower_extinction: lower limits for parameters of extinction model (keyword argument list)
        :param kwargs_upper_lens: upper limits for parameters of lens model (keyword argument list)
        :param kwargs_upper_source: upper limits for parameters of source model (keyword argument list)
        :param kwargs_upper_lens_light: upper limits for parameters of lens light model (keyword argument list)
        :param kwargs_upper_ps: upper limits for parameters of point source model (keyword argument list)
        :param kwargs_upper_special: upper limits for parameters of special model parameters (keyword arguments)
        :param kwargs_upper_extinction: upper limits for parameters of extinction model (keyword argument list)
        :param kwargs_lens_init: initial guess of lens model keyword arguments (only relevant as the starting point of
         the non-linear solver)
        :param linear_solver: bool, if True; avoids sampling the linear amplitude parameters 'amp' such that they
         get overwritten by the linear solver solution.
         Fixed 'amp' parameters will be overwritten if linear_solver = True.
        :param joint_lens_with_lens: list [[i_lens, k_lens, ['param_name1', 'param_name2', ...]], [...], ...],
         joint parameter between two lens models
        :param joint_lens_light_with_lens_light: list [[i_lens_light, k_lens_light, ['param_name1', 'param_name2', ...]], [...], ...],
         joint parameter between two lens light models, the second adopts the value of the first
        :param joint_source_with_source: [[i_source, k_source, ['param_name1', 'param_name2', ...]], [...], ...],
         joint parameter between two source surface brightness models, the second adopts the value of the first
        :param joint_lens_with_light: list [[i_light, k_lens, ['param_name1', 'param_name2', ...]], [...], ...],
         joint parameter between lens model and lens light model
        :param joint_source_with_point_source: list [[i_point_source, k_source], [...], ...],
         joint position parameter between lens model and source light model
        :param joint_lens_light_with_point_source: list [[i_point_source, k_lens_light], [...], ...],
         joint position parameter between lens model and lens light model
        :param joint_extinction_with_lens_light: list [[i_lens_light, k_extinction, ['param_name1', 'param_name2', ...]], [...], ...],
         joint parameters between the lens surface brightness and the optical depth models
        :param joint_lens_with_source_light: [[i_source, k_lens, ['param_name1', 'param_name2', ...]], [...], ...],
         joint parameter between lens model and source light model. Samples light model parameter only.
        :param mass_scaling_list: boolean list of length of lens model list (optional) models with identical integers
         will be scaled with the same additional scaling factor. First integer starts with 1 (not 0)
        :param general_scaling: { 'param_1': [list of booleans/integers defining which model to fit], 'param_2': [..], ..}
        :param point_source_offset: bool, if True, adds relative offsets ot the modeled image positions relative to the
         time-delay and lens equation solver
        :param num_point_source_list: list of number of point sources per point source model class
        :param image_plane_source_list: optional, list of booleans for the source_light components.
         If a component is set =True it will parameterized the positions in the image plane and ray-trace the
         parameters back to the source position on the fly during the fitting.
        :param solver_type: string, option for specific solver type
         see detailed instruction of the Solver4Point and Solver2Point classes
        :param Ddt_sampling: bool, if True, samples the time-delay distance D_dt (in units of Mpc)
        :param source_size: bool, if True, samples a source size parameters to be evaluated in the flux ratio likelihood
        :param num_tau0: integer, number of different optical depth re-normalization factors
        :param lens_redshift_sampling_indexes: list of integers corresponding to the lens model components whose redshifts
         are a free parameter (only has an effect in multi-plane lensing) with same indexes indicating joint redshift,
         in ascending numbering e.g. [-1, 0, 0, 1, 0, 2], -1 indicating not sampled fixed indexes
        :param source_redshift_sampling_indexes: list of integers corresponding to the source model components whose redshifts
         are a free parameter (only has an effect in multi-plane lensing) with same indexes indicating joint redshift,
         in ascending numbering e.g. [-1, 0, 0, 1, 0, 2], -1 indicating not sampled fixed indexes. These indexes are
         the sample as for the lens
        :param source_grid_offset: optional, if True when using a pixel-based modelling (e.g. with STARLETS-like profiles),
         adds two additional sampled parameters describing RA/Dec offsets between data coordinate grid and pixelated source plane coordinate grid.
        :param num_shapelet_lens: number of shapelet coefficients in the 'SHAPELETS_CART' or 'SHAPELETS_POLAR' mass profile.
        :param log_sampling_lens: Sample the log10 of the lens model parameters. Format : [[i_lens, ['param_name1', 'param_name2', ...]], [...], ...],
        """

        self._lens_model_list = kwargs_model.get("lens_model_list", [])
        self._lens_redshift_list = kwargs_model.get("lens_redshift_list", None)
        self._source_light_model_list = kwargs_model.get("source_light_model_list", [])
        self._source_redshift_list = kwargs_model.get("source_redshift_list", None)
        self._lens_light_model_list = kwargs_model.get("lens_light_model_list", [])
        self._point_source_model_list = kwargs_model.get("point_source_model_list", [])
        self._optical_depth_model_list = kwargs_model.get(
            "optical_depth_model_list", []
        )
        self._kwargs_model = kwargs_model

        # check how many redshifts need to be sampled
        num_z_sampling = 0
        if lens_redshift_sampling_indexes is not None:
            num_z_sampling = int(np.max(lens_redshift_sampling_indexes) + 1)
        if source_redshift_sampling_indexes is not None:
            num_z_source = int(np.max(source_redshift_sampling_indexes) + 1)
            num_z_sampling = max(num_z_sampling, num_z_source)
        (
            self._num_z_sampling,
            self._lens_redshift_sampling_indexes,
            self._source_redshift_sampling_indexes,
        ) = (
            num_z_sampling,
            lens_redshift_sampling_indexes,
            source_redshift_sampling_indexes,
        )

        (
            self._lens_model_class,
            self._source_model_class,
            _,
            _,
            _,
        ) = class_creator.create_class_instances(all_models=True, **kwargs_model)
        self._image2SourceMapping = Image2SourceMapping(
            lensModel=self._lens_model_class, sourceModel=self._source_model_class
        )

        if kwargs_fixed_lens is None:
            kwargs_fixed_lens = [{} for _ in range(len(self._lens_model_list))]
        if kwargs_fixed_source is None:
            kwargs_fixed_source = [
                {} for _ in range(len(self._source_light_model_list))
            ]
        if kwargs_fixed_lens_light is None:
            kwargs_fixed_lens_light = [
                {} for _ in range(len(self._lens_light_model_list))
            ]
        if kwargs_fixed_ps is None:
            kwargs_fixed_ps = [{} for _ in range(len(self._point_source_model_list))]
        if kwargs_fixed_special is None:
            kwargs_fixed_special = {}

        self._joint_lens_with_lens = joint_lens_with_lens
        self._joint_lens_light_with_lens_light = joint_lens_light_with_lens_light
        self._joint_source_with_source = joint_source_with_source

        self._joint_lens_with_light = joint_lens_with_light
        self._joint_lens_with_source_light = joint_lens_with_source_light
        self._joint_source_with_point_source = copy.deepcopy(
            joint_source_with_point_source
        )

        # Set up the parameters being sampled in log space in a similar way than the parameters being fixed.
        self._log_sampling_lens = log_sampling_lens
        kwargs_logsampling_lens = [[] for i in range(len(self._lens_model_list))]
        kwargs_logsampling_lens = self._update_log_sampling(
            kwargs_logsampling_lens, log_sampling_lens
        )

        for param_list in self._joint_source_with_point_source:
            if len(param_list) == 2:
                param_list.append(["center_x", "center_y"])
        self._joint_lens_light_with_point_source = copy.deepcopy(
            joint_lens_light_with_point_source
        )
        for param_list in self._joint_lens_light_with_point_source:
            if len(param_list) == 2:
                param_list.append(["center_x", "center_y"])
        if mass_scaling_list is None:
            mass_scaling_list = [False] * len(self._lens_model_list)
        self._mass_scaling_list = mass_scaling_list
        if 1 in self._mass_scaling_list:
            self._num_scale_factor = np.max(self._mass_scaling_list)
            self._mass_scaling = True
        else:
            self._num_scale_factor = 0
            self._mass_scaling = False

        if general_scaling is not None:
            self._general_scaling = True
            # FIXME TODO: check that the scaled parameters are actually used
            # by each lens model. For example, NFW has no theta_E parameter,
            # if a user tries to scale theta_E for an NFW we should throw an
            # error
            self._general_scaling_masks = dict(general_scaling)
        else:
            self._general_scaling = False
            self._general_scaling_masks = dict()

        self._point_source_offset = point_source_offset
        if num_point_source_list is None:
            num_point_source_list = [1] * len(self._point_source_model_list)

        # Attention: if joint coordinates with other source profiles, only indicate one as bool
        if image_plane_source_list is None:
            image_plane_source_list = [False] * len(self._source_light_model_list)
        self._image_plane_source_list = image_plane_source_list

        try:
            self._num_images = num_point_source_list[0]
        except:
            self._num_images = 0
        self._solver_type = solver_type
        if self._solver_type == "NONE":
            self._solver = False
        else:
            self._solver = True
            self._solver_module = Solver(
                solver_type=self._solver_type,
                lensModel=self._lens_model_class,
                num_images=self._num_images,
            )

        source_model_list = self._source_light_model_list
        if len(source_model_list) != 1 or source_model_list[0] not in [
            "SLIT_STARLETS",
            "SLIT_STARLETS_GEN2",
        ]:
            # source_grid_offset only defined for source profiles compatible with pixel-based solver
            source_grid_offset = False

        self._joint_extinction_with_lens_light = joint_extinction_with_lens_light
        # fix parameters joint within the same model types
        kwargs_fixed_lens_updated = self._add_fixed_lens(
            kwargs_fixed_lens, kwargs_lens_init
        )
        kwargs_fixed_lens_updated = self._fix_joint_param(
            kwargs_fixed_lens_updated, self._joint_lens_with_lens
        )
        kwargs_fixed_lens_updated = self._fix_joint_param(
            kwargs_fixed_lens_updated, self._joint_lens_with_source_light
        )
        kwargs_fixed_lens_light_updated = self._fix_joint_param(
            kwargs_fixed_lens_light, self._joint_lens_light_with_lens_light
        )
        kwargs_fixed_source_updated = self._fix_joint_param(
            kwargs_fixed_source, self._joint_source_with_source
        )
        kwargs_fixed_ps_updated = copy.deepcopy(kwargs_fixed_ps)
        kwargs_fixed_extinction_updated = self._fix_joint_param(
            kwargs_fixed_extinction, self._joint_extinction_with_lens_light
        )
        # fix parameters joint with other model types
        kwargs_fixed_lens_updated = self._fix_joint_param(
            kwargs_fixed_lens_updated, self._joint_lens_with_light
        )
        kwargs_fixed_source_updated = self._fix_joint_param(
            kwargs_fixed_source_updated, self._joint_source_with_point_source
        )
        kwargs_fixed_lens_light_updated = self._fix_joint_param(
            kwargs_fixed_lens_light_updated, self._joint_lens_light_with_point_source
        )
        self.lensParams = LensParam(
            self._lens_model_list,
            kwargs_fixed_lens_updated,
            kwargs_logsampling=kwargs_logsampling_lens,
            num_images=self._num_images,
            solver_type=self._solver_type,
            kwargs_lower=kwargs_lower_lens,
            kwargs_upper=kwargs_upper_lens,
            num_shapelet_lens=num_shapelet_lens,
        )
        self.lensLightParams = LightParam(
            self._lens_light_model_list,
            kwargs_fixed_lens_light_updated,
            param_type="lens_light",
            linear_solver=linear_solver,
            kwargs_lower=kwargs_lower_lens_light,
            kwargs_upper=kwargs_upper_lens_light,
        )
        self.sourceParams = LightParam(
            self._source_light_model_list,
            kwargs_fixed_source_updated,
            param_type="source_light",
            linear_solver=linear_solver,
            kwargs_lower=kwargs_lower_source,
            kwargs_upper=kwargs_upper_source,
        )
        self.pointSourceParams = PointSourceParam(
            self._point_source_model_list,
            kwargs_fixed_ps_updated,
            num_point_source_list=num_point_source_list,
            linear_solver=linear_solver,
            kwargs_lower=kwargs_lower_ps,
            kwargs_upper=kwargs_upper_ps,
        )
        self.extinctionParams = LightParam(
            self._optical_depth_model_list,
            kwargs_fixed_extinction_updated,
            kwargs_lower=kwargs_lower_extinction,
            kwargs_upper=kwargs_upper_extinction,
            linear_solver=False,
        )
        self.specialParams = SpecialParam(
            Ddt_sampling=Ddt_sampling,
            mass_scaling=self._mass_scaling,
            general_scaling_params=self._general_scaling_masks,
            kwargs_fixed=kwargs_fixed_special,
            num_scale_factor=self._num_scale_factor,
            kwargs_lower=kwargs_lower_special,
            kwargs_upper=kwargs_upper_special,
            point_source_offset=self._point_source_offset,
            num_images=self._num_images,
            source_size=source_size,
            num_tau0=num_tau0,
            num_z_sampling=num_z_sampling,
            source_grid_offset=source_grid_offset,
            kinematic_sampling=kinematic_sampling,
        )
        for lens_source_joint in self._joint_lens_with_source_light:
            i_source = lens_source_joint[0]
            if i_source in self._image_plane_source_list:
                raise ValueError(
                    "linking a source light model with a lens model AND simultaneously parameterizing the"
                    " source position in the image plane is not valid!"
                )
        self._linear_solver = linear_solver

    @property
    def num_point_source_images(self):
        """

        :return: total number of point source images
        """
        return self._num_images

    @property
    def linear_solver(self):
        """Boolean to state whether linear solver is activated or not.

        :return: boolean
        """
        return self._linear_solver

    def args2kwargs(self, args, bijective=False):
        """

        :param args: tuple of parameter values (float, strings, ...)
        :param bijective: boolean, if True (default) returns the parameters in the form as they are sampled
         (e.g. if image_plane_source_list is set =True it returns the position in the image plane coordinates),
         if False, returns the parameters in the form to render a model (e.g. image_plane_source_list positions are
         ray-traced back to the source plane).
        :return: keyword arguments sorted in lenstronomy conventions
        """
        i = 0
        args = np.atleast_1d(args)
        kwargs_lens, i = self.lensParams.get_params(args, i)
        kwargs_source, i = self.sourceParams.get_params(args, i)
        kwargs_lens_light, i = self.lensLightParams.get_params(args, i)
        kwargs_ps, i = self.pointSourceParams.get_params(args, i)
        kwargs_special, i = self.specialParams.get_params(args, i, impose_bound=True)
        kwargs_extinction, i = self.extinctionParams.get_params(args, i)
        self._update_lens_model(kwargs_special)
        # update lens_light joint parameters
        kwargs_lens_light = self._update_lens_light_joint_with_point_source(
            kwargs_lens_light, kwargs_ps
        )
        kwargs_lens_light = self._update_joint_param(
            kwargs_lens_light, kwargs_lens_light, self._joint_lens_light_with_lens_light
        )
        # update lens_light joint with lens model parameters
        kwargs_lens = self._update_joint_param(
            kwargs_lens_light, kwargs_lens, self._joint_lens_with_light
        )
        kwargs_lens = self._update_joint_param(
            kwargs_source, kwargs_lens, self._joint_lens_with_source_light
        )
        # update extinction model with lens light model
        kwargs_extinction = self._update_joint_param(
            kwargs_lens_light, kwargs_extinction, self._joint_extinction_with_lens_light
        )
        # update lens model joint parameters (including scaling)
        kwargs_lens = self._update_joint_param(
            kwargs_lens, kwargs_lens, self._joint_lens_with_lens
        )
        kwargs_lens = self.update_lens_scaling(kwargs_special, kwargs_lens)
        # update point source constraint solver
        if self._solver is True:
            x_pos, y_pos = kwargs_ps[0]["ra_image"], kwargs_ps[0]["dec_image"]
            kwargs_lens = self._solver_module.update_solver(kwargs_lens, x_pos, y_pos)
        # update source joint with point source
        kwargs_source = self._update_source_joint_with_point_source(
            kwargs_lens, kwargs_source, kwargs_ps, kwargs_special, image_plane=bijective
        )
        # update source joint with source
        kwargs_source = self._update_joint_param(
            kwargs_source, kwargs_source, self._joint_source_with_source
        )
        # optional revert lens_scaling for bijective
        if bijective is True:
            kwargs_lens = self.update_lens_scaling(
                kwargs_special, kwargs_lens, inverse=True
            )
        kwargs_return = {
            "kwargs_lens": kwargs_lens,
            "kwargs_source": kwargs_source,
            "kwargs_lens_light": kwargs_lens_light,
            "kwargs_ps": kwargs_ps,
            "kwargs_special": kwargs_special,
            "kwargs_extinction": kwargs_extinction,
        }
        return kwargs_return

    def kwargs2args(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_special=None,
        kwargs_extinction=None,
    ):
        """Inverse of getParam function :param kwargs_lens: keyword arguments depending
        on model options :param kwargs_source: keyword arguments depending on model
        options :param kwargs_lens_light: lens light model keyword argument list :param
        kwargs_ps: point source model keyword argument list :param kwargs_special:
        special keyword arguments :param kwargs_extinction: extinction model keyword
        argument list :return: numpy array of parameters."""
        args = self.lensParams.set_params(kwargs_lens)
        args += self.sourceParams.set_params(kwargs_source)
        args += self.lensLightParams.set_params(kwargs_lens_light)
        args += self.pointSourceParams.set_params(kwargs_ps)
        args += self.specialParams.set_params(kwargs_special)
        args += self.extinctionParams.set_params(kwargs_extinction)
        return np.array(args, dtype=float)

    def param_limits(self):
        """

        :return: lower and upper limits of the arguments being sampled
        """
        lower_limit = self.kwargs2args(
            kwargs_lens=self.lensParams.lower_limit,
            kwargs_source=self.sourceParams.lower_limit,
            kwargs_lens_light=self.lensLightParams.lower_limit,
            kwargs_ps=self.pointSourceParams.lower_limit,
            kwargs_special=self.specialParams.lower_limit,
            kwargs_extinction=self.extinctionParams.lower_limit,
        )
        upper_limit = self.kwargs2args(
            kwargs_lens=self.lensParams.upper_limit,
            kwargs_source=self.sourceParams.upper_limit,
            kwargs_lens_light=self.lensLightParams.upper_limit,
            kwargs_ps=self.pointSourceParams.upper_limit,
            kwargs_special=self.specialParams.upper_limit,
            kwargs_extinction=self.extinctionParams.upper_limit,
        )
        return lower_limit, upper_limit

    def num_param(self):
        """

        :return: number of parameters involved (int), list of parameter names
        """
        num, name_list = self.lensParams.num_param()
        _num, _list = self.sourceParams.num_param()
        num += _num
        name_list += _list
        _num, _list = self.lensLightParams.num_param()
        num += _num
        name_list += _list
        _num, _list = self.pointSourceParams.num_param()
        num += _num
        name_list += _list
        _num, _list = self.specialParams.num_param()
        num += _num
        name_list += _list
        _num, _list = self.extinctionParams.num_param()
        num += _num
        name_list += _list
        return num, name_list

    def num_param_linear(self):
        """

        :return: number of linear basis set coefficients that are solved for
        """
        num = 0
        num += self.sourceParams.num_param_linear()
        num += self.lensLightParams.num_param_linear()
        num += self.pointSourceParams.num_param_linear()
        return num

    def image2source_plane(self, kwargs_source, kwargs_lens, image_plane=False):
        """Maps the image plane position definition of the source plane.

        :param kwargs_source: source light model keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :param image_plane: boolean, if True, does not up map image plane parameters to
            source plane
        :return: source light model keyword arguments with mapped position arguments
            from image to source plane
        """
        kwargs_source_copy = copy.deepcopy(kwargs_source)
        for i, kwargs in enumerate(kwargs_source_copy):
            if self._image_plane_source_list[i] is True and not image_plane:
                if "center_x" in kwargs:
                    x_mapped, y_mapped = self._image2SourceMapping.image2source(
                        kwargs["center_x"],
                        kwargs["center_y"],
                        kwargs_lens,
                        index_source=i,
                    )
                    kwargs["center_x"] = x_mapped
                    kwargs["center_y"] = y_mapped
        return kwargs_source_copy

    def _update_source_joint_with_point_source(
        self,
        kwargs_lens_list,
        kwargs_source_list,
        kwargs_ps,
        kwargs_special,
        image_plane=False,
    ):
        kwargs_source_list = self.image2source_plane(
            kwargs_source_list, kwargs_lens_list, image_plane=image_plane
        )

        for setting in self._joint_source_with_point_source:
            i_point_source, k_source, param_list = setting
            if "ra_source" in kwargs_ps[i_point_source]:
                x_mapped = kwargs_ps[i_point_source]["ra_source"]
                y_mapped = kwargs_ps[i_point_source]["dec_source"]
            else:
                x_pos, y_pos = (
                    kwargs_ps[i_point_source]["ra_image"],
                    kwargs_ps[i_point_source]["dec_image"],
                )
                # x_pos, y_pos = self.real_image_positions(x_pos, y_pos, kwargs_special)
                x_mapped, y_mapped = self._image2SourceMapping.image2source(
                    x_pos, y_pos, kwargs_lens_list, index_source=k_source
                )
            for param_name in param_list:
                if param_name == "center_x":
                    kwargs_source_list[k_source][param_name] = np.mean(x_mapped)
                elif param_name == "center_y":
                    kwargs_source_list[k_source][param_name] = np.mean(y_mapped)
                else:
                    kwargs_source_list[k_source][param_name] = kwargs_ps[
                        i_point_source
                    ][param_name]
        return kwargs_source_list

    def _update_lens_light_joint_with_point_source(
        self, kwargs_lens_light_list, kwargs_ps
    ):
        for setting in self._joint_lens_light_with_point_source:
            i_point_source, k_lens_light, param_list = setting
            if "ra_image" in kwargs_ps[i_point_source]:
                x_mapped = kwargs_ps[i_point_source]["ra_image"]
                y_mapped = kwargs_ps[i_point_source]["dec_image"]
            else:
                raise ValueError(
                    "Joint lens light with point source not possible as point source is defined in the source plane!"
                )
            for param_name in param_list:
                if param_name == "center_x":
                    kwargs_lens_light_list[k_lens_light][param_name] = np.mean(x_mapped)
                elif param_name == "center_y":
                    kwargs_lens_light_list[k_lens_light][param_name] = np.mean(y_mapped)
        return kwargs_lens_light_list

    @staticmethod
    def _update_joint_param(kwargs_list_1, kwargs_list_2, joint_setting_list):
        """

        :param kwargs_list_1: list of keyword arguments
        :param kwargs_list_2: list of keyword arguments
        :param joint_setting_list: [[i_1, k_2, ['param_name1', 'param_name2', ...]], [...], ...]
                                     or: [[i_1, k_2, {'param_old1': 'param_new1', 'ra_0': 'center_x'}], [...], ...]
        :return: updated kwargs_list_2 with arguments from kwargs_list_1 as defined in joint_setting_list
        """
        for setting in joint_setting_list:
            i_1, k_2, param_list = setting
            if type(param_list) == list:
                for param_name in param_list:
                    kwargs_list_2[k_2][param_name] = kwargs_list_1[i_1][param_name]
            elif type(param_list) == dict:
                for param_to, param_from in param_list.items():
                    kwargs_list_2[k_2][param_to] = kwargs_list_1[i_1][param_from]
            else:
                raise TypeError(
                    "Bad format for constraint setting: got %s" % param_list
                )
        return kwargs_list_2

    @staticmethod
    def _update_log_sampling(kwargs_logsampling_lens, log_sampling_lens):
        """Update the list of parameters being sampled in log-space :param
        kwargs_logsampling_lens: list of list of parameters to sample in log10 :param
        log_sampling_lens: [[i_1, ['param_name1', 'param_name2', ...]], [...], ...]
        :return: updated kwargs_logsampling_lens."""
        for setting in log_sampling_lens:
            i_1, param_list = setting
            if type(param_list) == list:
                kwargs_logsampling_lens[i_1] = param_list
            else:
                raise TypeError(
                    "Bad format for constraint setting: got %s. This should be in the format [[i_1, ['param_name1', 'param_name2', ...]], [...], ...]"
                    % param_list
                )
        return kwargs_logsampling_lens

    @staticmethod
    def _fix_joint_param(kwargs_list_2, joint_setting_list):
        """

        :param kwargs_list_2: list of keyword arguments
        :param joint_setting_list: [[i_1, k_2, ['param_name1', 'param_name2', ...]], [...], ...]
        :return: fixes entries in kwargs_list_2 that are joint with other kwargs_list as defined in joint_setting_list
        """
        kwargs_list_2_update = copy.deepcopy(kwargs_list_2)
        for setting in joint_setting_list:
            i_1, k_2, param_list = setting
            for param_name in param_list:
                kwargs_list_2_update[k_2][param_name] = 0
        return kwargs_list_2_update

    def update_lens_scaling(self, kwargs_special, kwargs_lens, inverse=False):
        """Multiplies the scaling parameters of the profiles.

        :param kwargs_special: keyword arguments of the 'special' arguments
        :param kwargs_lens: lens model keyword argument list
        :param inverse: bool, if True, performs the inverse lens scaling for bijective
            transforms
        :return: updated lens model keyword argument list
        """
        kwargs_lens_updated = copy.deepcopy(kwargs_lens)
        # If we do not scaling, there's nothing to be done
        if not (self._mass_scaling or self._general_scaling):
            return kwargs_lens_updated

        # TODO: remove separate logic for mass scaling. either deprecate it
        # entirely, implement the details as a special case of general_scaling
        if self._mass_scaling:
            scale_factor_list = np.array(kwargs_special["scale_factor"])
            if inverse is True:
                scale_factor_list = 1.0 / np.array(kwargs_special["scale_factor"])
            for i, kwargs in enumerate(kwargs_lens_updated):
                if self._mass_scaling_list[i] is not False:
                    scale_factor = scale_factor_list[self._mass_scaling_list[i] - 1]
                    if "theta_E" in kwargs:
                        kwargs["theta_E"] *= scale_factor
                    elif "alpha_Rs" in kwargs:
                        kwargs["alpha_Rs"] *= scale_factor
                    elif "alpha_1" in kwargs:
                        kwargs["alpha_1"] *= scale_factor
                    elif "sigma0" in kwargs:
                        kwargs["sigma0"] *= scale_factor
                    elif "k_eff" in kwargs:
                        kwargs["k_eff"] *= scale_factor

        if self._general_scaling:
            for param_name in self._general_scaling_masks.keys():
                factors = kwargs_special[f"{param_name}_scale_factor"]
                _pows = kwargs_special[f"{param_name}_scale_pow"]

                for i, kwargs in enumerate(kwargs_lens_updated):
                    scale_idx = self._general_scaling_masks[param_name][i]
                    if scale_idx is not False:
                        if inverse:
                            kwargs[param_name] = (
                                kwargs[param_name] / factors[scale_idx - 1]
                            ) ** (1 / _pows[scale_idx - 1])
                        else:
                            kwargs[param_name] = (
                                factors[scale_idx - 1]
                                * kwargs[param_name] ** _pows[scale_idx - 1]
                            )

        return kwargs_lens_updated

    def _add_fixed_lens(self, kwargs_fixed, kwargs_init):
        kwargs_fixed_update = copy.deepcopy(kwargs_fixed)
        if self._solver is True:
            if kwargs_init is None:
                raise ValueError(
                    "kwargs_lens_init must be specified when the point source solver is enabled!"
                )
            kwargs_fixed_update = self._solver_module.add_fixed_lens(
                kwargs_fixed_update, kwargs_init
            )
        return kwargs_fixed_update

    def update_kwargs_model(self, kwargs_special):
        """Updates model keyword arguments with redshifts being sampled.

        :param kwargs_special: keyword arguments from SpecialParam() class return of
            sampling arguments
        :return: kwargs_model, bool (True if kwargs_model has changed, else False)
        """
        if self._num_z_sampling == 0:
            return self._kwargs_model, False
        z_samples = kwargs_special.get("z_sampling")
        lens_redshift_list = copy.deepcopy(self._lens_redshift_list)
        if not (
            self._lens_redshift_list is None
            or self._lens_redshift_sampling_indexes is None
        ):
            # iterate through index lists
            for i, index in enumerate(self._lens_redshift_sampling_indexes):
                # update redshifts of lens and source redshift list in new form
                if index > -1:
                    lens_redshift_list[i] = z_samples[index]
        source_redshift_list = copy.deepcopy(self._source_redshift_list)
        if not (
            self._source_redshift_list is None
            or self._source_redshift_sampling_indexes is None
        ):
            # iterate through index lists
            for i, index in enumerate(self._source_redshift_sampling_indexes):
                # update redshifts of lens and source redshift list in new form
                if index > -1:
                    source_redshift_list[i] = z_samples[index]
        # update lens model and source model classes
        kwargs_model = copy.deepcopy(self._kwargs_model)
        kwargs_model["lens_redshift_list"] = lens_redshift_list
        kwargs_model["source_redshift_list"] = source_redshift_list
        return kwargs_model, True

    def _update_lens_model(self, kwargs_special):
        """Updates lens model instance of this class (and all class instances related to
        it) when an update to the modeled redshifts of the deflector and/or source
        planes are made.

        :param kwargs_special: keyword arguments from SpecialParam() class return of
            sampling arguments
        :return: None, internal calls instance updated
        """
        kwargs_model, update_bool = self.update_kwargs_model(kwargs_special)
        if update_bool is True:
            # TODO: this class instances are effectively duplicated in the likelihood module and may cause a lot of overhead
            # in the calculation as the instances are re-generated every step, and even so doing it twice!
            (
                self._lens_model_class,
                self._source_model_class,
                _,
                _,
                _,
            ) = class_creator.create_class_instances(all_models=True, **kwargs_model)
            self._image2SourceMapping = Image2SourceMapping(
                lensModel=self._lens_model_class, sourceModel=self._source_model_class
            )

    def check_solver(self, kwargs_lens, kwargs_ps):
        """Test whether the image positions map back to the same source position :param
        kwargs_lens: lens model keyword argument list :param kwargs_ps: point source
        model keyword argument list :return: Euclidean distance between the ray-shooting
        of the image positions."""
        if self._solver is True:
            image_x, image_y = kwargs_ps[0]["ra_image"], kwargs_ps[0]["dec_image"]
            dist = self._solver_module.check_solver(image_x, image_y, kwargs_lens)
            return np.max(dist)
        else:
            return 0

    def print_setting(self):
        """Prints the setting of the parameter class.

        :return:
        """
        num, param_list = self.num_param()
        num_linear = self.num_param_linear()

        # TODO print settings of specailParams?
        print("The following model options are chosen:")
        print("Lens models:", self._lens_model_list)
        print("Source models:", self._source_light_model_list)
        print("Lens light models:", self._lens_light_model_list)
        print("Point source models:", self._point_source_model_list)
        print("===================")
        print("The following parameters are being fixed:")
        print("Lens:", self.lensParams.kwargs_fixed)
        print("Source:", self.sourceParams.kwargs_fixed)
        print("Lens light:", self.lensLightParams.kwargs_fixed)
        print("Point source:", self.pointSourceParams.kwargs_fixed)
        print("===================")
        print("Joint parameters for different models")
        print("Joint lens with lens:", self._joint_lens_with_lens)
        print(
            "Joint lens light with lens light:", self._joint_lens_light_with_lens_light
        )
        print("Joint source with source:", self._joint_source_with_source)
        print("Joint lens with light:", self._joint_lens_with_light)
        print("Joint source with point source:", self._joint_source_with_point_source)
        print(
            "Joint lens light with point source:",
            self._joint_lens_light_with_point_source,
        )
        print("Mass scaling:", self._num_scale_factor, "groups")
        print("General lens scaling:", self._general_scaling_masks)
        print("===================")
        print("Number of non-linear parameters being sampled: ", num)
        print("Parameters being sampled: ", param_list)
        print("Number of linear parameters being solved for: ", num_linear)
        print("===================")
        print("The log10 of following parameters is being sampled:")
        print("Lens:", self.lensParams.kwargs_logsampling)



# conftest.py
import pytest

import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.Sampling.likelihood import LikelihoodModule
from lenstronomy.Sampling.parameters import Param
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF


@pytest.fixture
def simple_einstein_ring_likelihood():
    # data specifics
    sigma_bkg = 0.05  # background noise per pixel
    exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
    numPix = 10  # cutout pixel size
    deltaPix = 0.1  # pixel size in arcsec (area per pixel = deltaPix**2)
    fwhm = 0.5  # full width half max of PSF

    # PSF specification

    kwargs_data = sim_util.data_configure_simple(numPix, deltaPix, exp_time, sigma_bkg)
    data_class = ImageData(**kwargs_data)
    kwargs_psf_gaussian = {"psf_type": "GAUSSIAN", "fwhm": fwhm, "pixel_size": deltaPix}
    psf = PSF(**kwargs_psf_gaussian)
    kwargs_psf = {"psf_type": "PIXEL", "kernel_point_source": psf.kernel_point_source}
    psf_class = PSF(**kwargs_psf)
    kwargs_spemd = {
        "theta_E": 1.0,
        "gamma": 1.8,
        "center_x": 0,
        "center_y": 0,
        "e1": 0.1,
        "e2": 0.1,
    }

    lens_model_list = ["SPEP"]
    kwargs_lens = [kwargs_spemd]
    lens_model_class = LensModel(lens_model_list=lens_model_list)
    kwargs_sersic = {
        "amp": 1.0,
        "R_sersic": 0.1,
        "n_sersic": 2,
        "center_x": 0,
        "center_y": 0,
    }
    # 'SERSIC_ELLIPSE': elliptical Sersic profile
    kwargs_sersic_ellipse = {
        "amp": 1.0,
        "R_sersic": 0.6,
        "n_sersic": 3,
        "center_x": 0,
        "center_y": 0,
        "e1": 0.1,
        "e2": 0.1,
    }

    lens_light_model_list = ["SERSIC"]
    kwargs_lens_light = [kwargs_sersic]
    lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
    source_model_list = ["SERSIC_ELLIPSE"]
    kwargs_source = [kwargs_sersic_ellipse]
    source_model_class = LightModel(light_model_list=source_model_list)

    kwargs_numerics = {
        "supersampling_factor": 1,
        "supersampling_convolution": False,
        "compute_mode": "regular",
    }
    imageModel = ImageModel(
        data_class,
        psf_class,
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        kwargs_numerics=kwargs_numerics,
    )
    image_sim = sim_util.simulate_simple(
        imageModel, kwargs_lens, kwargs_source, kwargs_lens_light
    )

    data_class.update_data(image_sim)
    kwargs_data["image_data"] = image_sim
    kwargs_data_joint = {
        "multi_band_list": [[kwargs_data, kwargs_psf, kwargs_numerics]],
        "multi_band_type": "single-band",
    }

    kwargs_model = {
        "lens_model_list": lens_model_list,
        "source_light_model_list": source_model_list,
        "lens_light_model_list": lens_light_model_list,
        "fixed_magnification_list": [False],
    }

    kwargs_constraints = {"image_plane_source_list": [False] * len(source_model_list)}

    kwargs_likelihood = {
        "source_marg": False,
        "image_position_uncertainty": 0.004,
        "check_matched_source_position": False,
        "source_position_tolerance": 0.001,
        "source_position_sigma": 0.001,
    }

    # reduce number of param to sample (for runtime)
    kwargs_fixed_lens = [
        {"gamma": 1.8, "center_x": 0, "center_y": 0, "e1": 0.1, "e2": 0.1}
    ]
    kwargs_lower_lens = [{"theta_E": 0.8, "center_x": -0.1}]
    kwargs_upper_lens = [{"theta_E": 1.2, "center_x": 0.1}]
    kwargs_fixed_source = kwargs_source
    kwargs_fixed_lens_light = kwargs_lens_light

    param_class = Param(
        kwargs_model,
        kwargs_fixed_lens=kwargs_fixed_lens,
        kwargs_fixed_source=kwargs_fixed_source,
        kwargs_fixed_lens_light=kwargs_fixed_lens_light,
        kwargs_lower_lens=kwargs_lower_lens,
        kwargs_upper_lens=kwargs_upper_lens,
        **kwargs_constraints
    )

    likelihood = LikelihoodModule(
        kwargs_data_joint=kwargs_data_joint,
        kwargs_model=kwargs_model,
        param_class=param_class,
        **kwargs_likelihood
    )
    kwargs_truths = {
        "kwargs_lens": kwargs_lens,
        "kwargs_source": kwargs_source,
        "kwargs_lens_light": kwargs_lens_light,
    }
    return likelihood, kwargs_truths


@pytest.fixture
def simple_einstein_ring_likelihood_2d():
    # data specifics
    sigma_bkg = 0.05  # background noise per pixel
    exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
    numPix = 10  # cutout pixel size
    deltaPix = 0.1  # pixel size in arcsec (area per pixel = deltaPix**2)
    fwhm = 0.5  # full width half max of PSF

    # PSF specification

    kwargs_data = sim_util.data_configure_simple(numPix, deltaPix, exp_time, sigma_bkg)
    data_class = ImageData(**kwargs_data)
    kwargs_psf_gaussian = {"psf_type": "GAUSSIAN", "fwhm": fwhm, "pixel_size": deltaPix}
    psf = PSF(**kwargs_psf_gaussian)
    kwargs_psf = {"psf_type": "PIXEL", "kernel_point_source": psf.kernel_point_source}
    psf_class = PSF(**kwargs_psf)
    kwargs_spemd = {
        "theta_E": 1.0,
        "gamma": 1.8,
        "center_x": 0,
        "center_y": 0,
        "e1": 0.1,
        "e2": 0.1,
    }

    lens_model_list = ["SPEP"]
    kwargs_lens = [kwargs_spemd]
    lens_model_class = LensModel(lens_model_list=lens_model_list)
    kwargs_sersic = {
        "amp": 1.0,
        "R_sersic": 0.1,
        "n_sersic": 2,
        "center_x": 0,
        "center_y": 0,
    }
    # 'SERSIC_ELLIPSE': elliptical Sersic profile
    kwargs_sersic_ellipse = {
        "amp": 1.0,
        "R_sersic": 0.6,
        "n_sersic": 3,
        "center_x": 0,
        "center_y": 0,
        "e1": 0.1,
        "e2": 0.1,
    }

    lens_light_model_list = ["SERSIC"]
    kwargs_lens_light = [kwargs_sersic]
    lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
    source_model_list = ["SERSIC_ELLIPSE"]
    kwargs_source = [kwargs_sersic_ellipse]
    source_model_class = LightModel(light_model_list=source_model_list)

    kwargs_numerics = {
        "supersampling_factor": 1,
        "supersampling_convolution": False,
        "compute_mode": "regular",
    }
    imageModel = ImageModel(
        data_class,
        psf_class,
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        kwargs_numerics=kwargs_numerics,
    )
    image_sim = sim_util.simulate_simple(
        imageModel, kwargs_lens, kwargs_source, kwargs_lens_light
    )

    data_class.update_data(image_sim)
    kwargs_data["image_data"] = image_sim
    kwargs_data_joint = {
        "multi_band_list": [[kwargs_data, kwargs_psf, kwargs_numerics]],
        "multi_band_type": "single-band",
    }

    kwargs_model = {
        "lens_model_list": lens_model_list,
        "source_light_model_list": source_model_list,
        "lens_light_model_list": lens_light_model_list,
        "fixed_magnification_list": [False],
    }

    kwargs_constraints = {"image_plane_source_list": [False] * len(source_model_list)}

    kwargs_likelihood = {
        "source_marg": False,
        "image_position_uncertainty": 0.004,
        "check_matched_source_position": False,
        "source_position_tolerance": 0.001,
        "source_position_sigma": 0.001,
    }

    # reduce number of param to sample (for runtime)
    kwargs_fixed_lens = [{"gamma": 1.8, "center_y": 0, "e1": 0.1, "e2": 0.1}]
    kwargs_lower_lens = [{"theta_E": 0.8, "center_x": -0.1}]
    kwargs_upper_lens = [{"theta_E": 1.2, "center_x": 0.1}]
    kwargs_fixed_source = [
        {
            "R_sersic": 0.6,
            "n_sersic": 3,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.1,
            "e2": 0.1,
        }
    ]
    kwargs_fixed_lens_light = [
        {"R_sersic": 0.1, "n_sersic": 2, "center_x": 0, "center_y": 0}
    ]

    param_class = Param(
        kwargs_model,
        kwargs_fixed_lens=kwargs_fixed_lens,
        kwargs_fixed_source=kwargs_fixed_source,
        kwargs_fixed_lens_light=kwargs_fixed_lens_light,
        kwargs_lower_lens=kwargs_lower_lens,
        kwargs_upper_lens=kwargs_upper_lens,
        **kwargs_constraints
    )

    likelihood = LikelihoodModule(
        kwargs_data_joint=kwargs_data_joint,
        kwargs_model=kwargs_model,
        param_class=param_class,
        **kwargs_likelihood
    )
    kwargs_truths = {
        "kwargs_lens": kwargs_lens,
        "kwargs_source": kwargs_source,
        "kwargs_lens_light": kwargs_lens_light,
    }
    return likelihood, kwargs_truths



__author__ = "sibirrer"

import lenstronomy.Util.param_util as param_util

import scipy.optimize
import numpy as np
import copy

__all__ = ["Solver4Point"]


class Solver4Point(object):
    """Class to make the constraints for the solver."""

    def __init__(self, lensModel, solver_type="PROFILE"):
        self._solver_type = solver_type  # supported:
        if not lensModel.lens_model_list[0] in [
            "SPEP",
            "SPEMD",
            "PEMD",
            "SIE",
            "NIE",
            "NFW_ELLIPSE",
            "NFW_ELLIPSE_CSE",
            "SHAPELETS_CART",
            "CNFW_ELLIPSE",
            "EPL",
            "EPL_BOXYDISKY",
        ]:
            raise ValueError(
                "first lens model must be supported by the solver: 'SPEP', 'SPEMD', 'PEMD',"
                " 'SIE', 'NIE', 'EPL', 'EPL_BOXYDISKY', 'NFW_ELLIPSE', 'NFW_ELLIPSE_CSE', "
                "'SHAPELETS_CART', 'CNFW_ELLIPSE'."
                "Your choice was %s" % lensModel.lens_model_list[0]
            )
        if solver_type not in ["PROFILE", "PROFILE_SHEAR"]:
            raise ValueError(
                "solver_type %s not supported! Choose from 'PROFILE', 'PROFILE_SHEAR'"
                % solver_type
            )
        if solver_type in ["PROFILE_SHEAR"]:
            if lensModel.lens_model_list[1] == "SHEAR":
                self._solver_type = "PROFILE_SHEAR"
            elif lensModel.lens_model_list[1] == "SHEAR_GAMMA_PSI":
                self._solver_type = "PROFILE_SHEAR_GAMMA_PSI"
            else:
                raise ValueError(
                    "second lens model must be SHEAR_GAMMA_PSI or SHEAR to enable solver type %s!"
                    % solver_type
                )
        self.lensModel = lensModel
        self._lens_mode_list = lensModel.lens_model_list
        if lensModel.multi_plane is True or "FOREGROUND_SHEAR" in self._lens_mode_list:
            self._decoupling = False
        else:
            self._decoupling = True

    def constraint_lensmodel(self, x_pos, y_pos, kwargs_list, xtol=1.49012e-12):
        """

        :param x_pos: list of image positions (x-axis)
        :param y_pos: list of image position (y-axis)
        :param xtol: numerical tolerance level
        :param kwargs_list: list of lens model kwargs
        :return: updated lens model that satisfies the lens equation for the point sources, accuracy
        """
        kwargs = copy.deepcopy(kwargs_list)
        init = self._extract_array(kwargs)
        if self._decoupling:
            alpha_0_x, alpha_0_y = self.lensModel.alpha(x_pos, y_pos, kwargs)
            alpha_1_x, alpha_1_y = self.lensModel.alpha(x_pos, y_pos, kwargs, k=0)
            if self._solver_type in ["PROFILE_SHEAR", "PROFILE_SHEAR_GAMMA_PSI"]:
                alpha_shear_x, alpha_shear_y = self.lensModel.alpha(
                    x_pos, y_pos, kwargs, k=1
                )
                alpha_1_x += alpha_shear_x
                alpha_1_y += alpha_shear_y
            x_sub = alpha_1_x - alpha_0_x
            y_sub = alpha_1_y - alpha_0_y
        else:
            x_sub, y_sub = np.zeros(4), np.zeros(4)
        a = self._subtract_constraint(x_sub, y_sub)
        x = self.solve(x_pos, y_pos, init, kwargs, a, xtol)
        kwargs = self._update_kwargs(x, kwargs)
        y_end = self._F(x, x_pos, y_pos, kwargs, a)
        accuracy = np.sum(y_end**2)
        return kwargs, accuracy

    def solve(self, x_pos, y_pos, init, kwargs_list, a, xtol=1.49012e-10):
        x = scipy.optimize.fsolve(
            self._F, init, args=(x_pos, y_pos, kwargs_list, a), xtol=xtol
        )  # , factor=0.1)
        return x

    def _F(self, x, x_pos, y_pos, kwargs_list, a=np.zeros(6)):
        kwargs_list = self._update_kwargs(x, kwargs_list)
        if self._decoupling:
            alpha_x, alpha_y = self.lensModel.alpha(x_pos, y_pos, kwargs_list, k=0)
            if self._solver_type in ["PROFILE_SHEAR", "PROFILE_SHEAR_GAMMA_PSI"]:
                alpha_x_shear, alpha_y_shear = self.lensModel.alpha(
                    x_pos, y_pos, kwargs_list, k=1
                )
                alpha_x += alpha_x_shear
                alpha_y += alpha_y_shear
            beta_x = x_pos - alpha_x
            beta_y = y_pos - alpha_y
        else:
            beta_x, beta_y = self.lensModel.ray_shooting(x_pos, y_pos, kwargs_list)
        y = np.zeros(6)
        y[0] = beta_x[0] - beta_x[1]
        y[1] = beta_x[0] - beta_x[2]
        y[2] = beta_x[0] - beta_x[3]
        y[3] = beta_y[0] - beta_y[1]
        y[4] = beta_y[0] - beta_y[2]
        y[5] = beta_y[0] - beta_y[3]
        return y - a

    @staticmethod
    def _subtract_constraint(x_sub, y_sub):
        """

        :param x_sub:
        :param y_sub:
        :return:
        """
        a = np.zeros(6)

        a[0] = -x_sub[0] + x_sub[1]
        a[1] = -x_sub[0] + x_sub[2]
        a[2] = -x_sub[0] + x_sub[3]
        a[3] = -y_sub[0] + y_sub[1]
        a[4] = -y_sub[0] + y_sub[2]
        a[5] = -y_sub[0] + y_sub[3]
        return a

    def _update_kwargs(self, x, kwargs_list):
        """

        :param x: list of parameters corresponding to the free parameter of the first lens model in the list
        :param kwargs_list: list of lens model kwargs
        :return: updated kwargs_list
        """
        if self._solver_type == "PROFILE_SHEAR_GAMMA_PSI":
            phi_G = x[5]  # % (2 * np.pi)
            kwargs_list[1]["psi_ext"] = phi_G
        if self._solver_type == "PROFILE_SHEAR":
            phi_G = x[5] % np.pi
            phi_G_no_sense, gamma_ext = param_util.shear_cartesian2polar(
                kwargs_list[1]["gamma1"], kwargs_list[1]["gamma2"]
            )
            gamma1, gamma2 = param_util.shear_polar2cartesian(phi_G, gamma_ext)
            kwargs_list[1]["gamma1"] = gamma1
            kwargs_list[1]["gamma2"] = gamma2
        lens_model = self._lens_mode_list[0]
        if lens_model in [
            "SPEP",
            "SPEMD",
            "SIE",
            "NIE",
            "PEMD",
            "EPL",
            "EPL_BOXYDISKY",
        ]:
            [theta_E, e1, e2, center_x, center_y, _] = x
            kwargs_list[0]["theta_E"] = theta_E
            kwargs_list[0]["e1"] = e1
            kwargs_list[0]["e2"] = e2
            kwargs_list[0]["center_x"] = center_x
            kwargs_list[0]["center_y"] = center_y

        elif lens_model in ["NFW_ELLIPSE", "CNFW_ELLIPSE", "NFW_ELLIPSE_CSE"]:
            [alpha_Rs, e1, e2, center_x, center_y, _] = x
            kwargs_list[0]["alpha_Rs"] = alpha_Rs
            kwargs_list[0]["e1"] = e1
            kwargs_list[0]["e2"] = e2
            kwargs_list[0]["center_x"] = center_x
            kwargs_list[0]["center_y"] = center_y
        elif lens_model in ["SHAPELETS_CART"]:
            [c10, c01, c20, c11, c02, _] = x
            coeffs = list(kwargs_list[0]["coeffs"])
            coeffs[1:6] = [c10, c01, c20, c11, c02]
            kwargs_list[0]["coeffs"] = coeffs
        else:
            raise ValueError(
                "Lens model %s not supported for 4-point solver!" % lens_model
            )
        return kwargs_list

    def _extract_array(self, kwargs_list):
        """Inverse of _update_kwargs :param kwargs_list:

        :return:
        """
        if self._solver_type == "PROFILE_SHEAR_GAMMA_PSI":
            phi_ext = kwargs_list[1]["psi_ext"]  # % (np.pi)
            # e1 = kwargs_list[1]['e1']
            # e2 = kwargs_list[1]['e2']
            # phi_ext, gamma_ext = param_util.ellipticity2phi_gamma(e1, e2)
        elif self._solver_type == "PROFILE_SHEAR":
            gamma1 = kwargs_list[1]["gamma1"]
            gamma2 = kwargs_list[1]["gamma2"]
            phi_ext, gamma_ext = param_util.shear_cartesian2polar(gamma1, gamma2)
            # phi_G_no_sense, gamma_ext = param_util.ellipticity2phi_gamma(kwargs_list[1]['e1'], kwargs_list[1]['e2'])
            # e1, e2 = param_util.phi_gamma_ellipticity(phi_G, gamma_ext)
            # kwargs_list[1]['e1'] = e1
        else:
            phi_ext = 0
        lens_model = self._lens_mode_list[0]
        if lens_model in [
            "SPEP",
            "SPEMD",
            "SIE",
            "NIE",
            "PEMD",
            "EPL",
            "EPL_BOXYDISKY",
        ]:
            e1 = kwargs_list[0]["e1"]
            e2 = kwargs_list[0]["e2"]
            center_x = kwargs_list[0]["center_x"]
            center_y = kwargs_list[0]["center_y"]
            theta_E = kwargs_list[0]["theta_E"]
            x = [theta_E, e1, e2, center_x, center_y, phi_ext]

        elif lens_model in ["NFW_ELLIPSE", "CNFW_ELLIPSE", "NFW_ELLIPSE_CSE"]:
            e1 = kwargs_list[0]["e1"]
            e2 = kwargs_list[0]["e2"]
            center_x = kwargs_list[0]["center_x"]
            center_y = kwargs_list[0]["center_y"]
            alpha_Rs = kwargs_list[0]["alpha_Rs"]
            x = [alpha_Rs, e1, e2, center_x, center_y, phi_ext]
        elif lens_model in ["SHAPELETS_CART"]:
            coeffs = list(kwargs_list[0]["coeffs"])
            [c10, c01, c20, c11, c02] = coeffs[1:6]
            x = [c10, c01, c20, c11, c02, phi_ext]
        else:
            raise ValueError(
                "Lens model %s not supported for 4-point solver!" % lens_model
            )
        return x

    def add_fixed_lens(self, kwargs_fixed_lens_list, kwargs_lens_init):
        """

        :param kwargs_fixed_lens_list:
        :param kwargs_lens_init:
        :return:
        """

        lens_model = self.lensModel.lens_model_list[0]
        kwargs_fixed = kwargs_fixed_lens_list[0]
        kwargs_lens = kwargs_lens_init[0]
        if self._solver_type in ["PROFILE_SHEAR", "PROFILE_SHEAR_GAMMA_PSI"]:
            pass
            # kwargs_fixed_lens_list[1]['psi_ext'] = kwargs_lens_init[1]['psi_ext']
        if lens_model in [
            "SPEP",
            "SPEMD",
            "SIE",
            "NIE",
            "PEMD",
            "EPL",
            "EPL_BOXYDISKY",
        ]:
            kwargs_fixed["theta_E"] = kwargs_lens["theta_E"]
            kwargs_fixed["e1"] = kwargs_lens["e1"]
            kwargs_fixed["e2"] = kwargs_lens["e2"]
            kwargs_fixed["center_x"] = kwargs_lens["center_x"]
            kwargs_fixed["center_y"] = kwargs_lens["center_y"]
        elif lens_model in ["NFW_ELLIPSE", "CNFW_ELLIPSE", "NFW_ELLIPSE_CSE"]:
            kwargs_fixed["alpha_Rs"] = kwargs_lens["alpha_Rs"]
            kwargs_fixed["e1"] = kwargs_lens["e1"]
            kwargs_fixed["e2"] = kwargs_lens["e2"]
            kwargs_fixed["center_x"] = kwargs_lens["center_x"]
            kwargs_fixed["center_y"] = kwargs_lens["center_y"]
        elif lens_model in ["SHAPELETS_CART"]:
            pass
        else:
            raise ValueError(
                "%s is not a valid option. Choose from 'PROFILE', 'PROFILE_SHEAR', 'SHAPELETS'"
                % self._solver_type
            )
        return kwargs_fixed_lens_list



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest

from lenstronomy.Cosmo.lens_cosmo import LensCosmo
from lenstronomy.Util import util


class TestLensCosmo(object):
    """Tests the UnitManager class routines."""

    def setup_method(self):
        z_L = 0.8
        z_S = 3.0
        from astropy.cosmology import FlatLambdaCDM

        cosmo = FlatLambdaCDM(H0=70, Om0=0.3, Ob0=0.05)
        self.lensCosmo = LensCosmo(z_L, z_S, cosmo=cosmo)

    def test_ang_dist(self):
        npt.assert_almost_equal(self.lensCosmo.ds, 1588.9213590743666, decimal=8)
        npt.assert_almost_equal(self.lensCosmo.dd, 1548.7055203661785, decimal=8)
        npt.assert_almost_equal(self.lensCosmo.dds, 892.0038749095863, decimal=8)

    def test_epsilon_crit(self):
        npt.assert_almost_equal(self.lensCosmo.sigma_crit / 1.9121e15, 1, decimal=3)

    def test_arcsec2phys(self):
        arcsec = np.array([1, 2])  # pixel coordinate from center
        physcoord = self.lensCosmo.arcsec2phys_lens(arcsec)
        npt.assert_almost_equal(physcoord[0], 0.0075083362428338641, decimal=8)
        npt.assert_almost_equal(physcoord[1], 0.015016672485667728, decimal=8)

        physcoord = self.lensCosmo.arcsec2phys_source(arcsec)
        npt.assert_almost_equal(physcoord[0], 0.007703308130864105, decimal=8)
        npt.assert_almost_equal(physcoord[1], 0.01540661626172821, decimal=8)

    def test_phys2arcsec_lens(self):
        phys = 1.0
        arc_sec = self.lensCosmo.phys2arcsec_lens(phys)
        phys_new = self.lensCosmo.arcsec2phys_lens(arc_sec)
        npt.assert_almost_equal(phys_new, phys, decimal=8)

    def test_mass_in_phi_E(self):
        phi_E = 1.5
        mass = self.lensCosmo.mass_in_theta_E(phi_E)
        npt.assert_almost_equal(mass, 761967261292.6725, decimal=2)

    def test_kappa2proj_mass(self):
        kappa = 0.5
        mass = self.lensCosmo.kappa2proj_mass(kappa)
        npt.assert_almost_equal(mass, kappa * self.lensCosmo.sigma_crit, decimal=3)

    def test_mass_in_coin(self):
        theta_E = 1.0
        m_coin = self.lensCosmo.mass_in_coin(theta_E)
        npt.assert_almost_equal(m_coin, 165279526936.52194, decimal=0)

    def test_D_dt_model(self):
        D_dt = self.lensCosmo.ddt
        npt.assert_almost_equal(D_dt, 4965.660384441859, decimal=8)

    def test_nfw_angle2physical(self):
        Rs_angle = 6.0
        alpha_Rs = 1.0
        rho0, Rs, c, r200, M200 = self.lensCosmo.nfw_angle2physical(Rs_angle, alpha_Rs)
        assert Rs * c == r200

    def test_nfw_physical2angle(self):
        M = 10.0**13.5
        c = 4
        Rs_angle, alpha_Rs = self.lensCosmo.nfw_physical2angle(M, c)
        rho0, Rs, c_out, r200, M200 = self.lensCosmo.nfw_angle2physical(
            Rs_angle, alpha_Rs
        )
        npt.assert_almost_equal(c_out, c, decimal=3)
        npt.assert_almost_equal(np.log10(M200), np.log10(M), decimal=4)

    def test_sis_theta_E2sigma_v(self):
        theta_E = 2.0
        sigma_v = self.lensCosmo.sis_theta_E2sigma_v(theta_E)
        theta_E_out = self.lensCosmo.sis_sigma_v2theta_E(sigma_v)
        npt.assert_almost_equal(theta_E_out, theta_E, decimal=5)

    def test_fermat2delays(self):
        fermat_pot = 0.5
        dt_days = self.lensCosmo.time_delay_units(fermat_pot)
        fermat_pot_out = self.lensCosmo.time_delay2fermat_pot(dt_days)
        npt.assert_almost_equal(fermat_pot, fermat_pot_out, decimal=10)

    def test_uldm_angular2phys(self):
        kappa_0, theta_c = 0.1, 3
        mlog10, Mlog10 = self.lensCosmo.uldm_angular2phys(kappa_0, theta_c)
        npt.assert_almost_equal(mlog10, -24.3610006, decimal=5)
        npt.assert_almost_equal(Mlog10, 11.7195843, decimal=5)

    def test_uldm_mphys2angular(self):
        m_log10, M_log10 = -24, 11
        kappa_0, theta_c = self.lensCosmo.uldm_mphys2angular(m_log10, M_log10)
        mcheck, Mcheck = self.lensCosmo.uldm_angular2phys(kappa_0, theta_c)
        npt.assert_almost_equal(mcheck, m_log10, decimal=4)
        npt.assert_almost_equal(Mcheck, M_log10, decimal=4)

    def test_a_z(self):
        a = self.lensCosmo.background.a_z(z=1)
        npt.assert_almost_equal(a, 0.5)

    def test_sersic_m_star2k_eff(self):
        m_star = 10**11.5
        R_sersic = 1
        n_sersic = 4
        k_eff = self.lensCosmo.sersic_m_star2k_eff(m_star, R_sersic, n_sersic)
        npt.assert_almost_equal(k_eff, 0.1294327891669961, decimal=5)

        m_star_out = self.lensCosmo.sersic_k_eff2m_star(k_eff, R_sersic, n_sersic)
        npt.assert_almost_equal(m_star_out, m_star, decimal=6)

    def test_hernquist_angular2phys(self):
        m_star = 10**10  # in M_sun
        rs = 0.01  # in Mpc

        # test bijective transformation
        sigma0, rs_angle = self.lensCosmo.hernquist_phys2angular(mass=m_star, rs=rs)
        m_star_new, rs_new = self.lensCosmo.hernquist_angular2phys(
            sigma0=sigma0, rs_angle=rs_angle
        )
        npt.assert_almost_equal(m_star_new, m_star, decimal=1)
        npt.assert_almost_equal(rs_new, rs, decimal=8)

    def test_hernquist_mass_normalization(self):
        m_star = 10**10  # in M_sun
        rs = 0.01  # in Mpc

        # test bijective transformation
        sigma0, rs_angle = self.lensCosmo.hernquist_phys2angular(mass=m_star, rs=rs)

        # test mass integrals
        # make large grid
        delta_pix = rs_angle / 30.0
        x, y = util.make_grid(numPix=501, deltapix=delta_pix)
        # compute convergence
        from lenstronomy.LensModel.lens_model import LensModel

        lens_model = LensModel(lens_model_list=["HERNQUIST"])
        kwargs = [{"sigma0": sigma0, "Rs": rs_angle, "center_x": 0, "center_y": 0}]
        kappa = lens_model.kappa(x, y, kwargs)
        # sum up convergence
        kappa_tot = np.sum(kappa) * delta_pix**2
        # transform to mass
        mass_tot = kappa_tot * self.lensCosmo.sigma_crit_angle

        # compare
        npt.assert_almost_equal(mass_tot / m_star, 1, decimal=1)


if __name__ == "__main__":
    pytest.main()



from coolest.template.json import (
    JSONSerializer,
)  # install from https://github.com/aymgal/COOLEST
from astropy.cosmology import FlatLambdaCDM
from astropy.io import fits
import lenstronomy.Util.coolest_read_util as read
import lenstronomy.Util.coolest_update_util as update
import numpy as np
from lenstronomy.Sampling.parameters import Param
import lenstronomy.Util.class_creator as class_util


def create_lenstronomy_from_coolest(file_name):
    """Creates lenstronomy typical kwargs from a COOLEST (JSON) file.

    :param file_name: str, name (with path) of the .json file containing the COOLEST
        information
    :return: return_dict, dictionary with typical lenstronomy kwarg (as kwargs_data,
        kwargs_psf, kwargs_params, kwargs_results, kwargs_model etc)
    """
    creation_lens_source_light = False
    creation_cosmo = False
    creation_data = False
    creation_instrument = False
    creation_redshift_list = False
    creation_kwargs_likelihood = False

    decoder = JSONSerializer(file_name, indent=2)
    lens_coolest = decoder.load()

    print(f"LENS COOLEST : {lens_coolest.mode}")

    # IMAGE

    kwargs_data = {}
    if lens_coolest.observation is not None:
        lens_observation = lens_coolest.observation
        if lens_observation.pixels is not None:
            creation_data = True
            image_path = lens_observation.pixels.fits_file.path
            pixel_size = lens_observation.pixels.pixel_size
            nx = lens_observation.pixels.num_pix_x
            ny = lens_observation.pixels.num_pix_y
            try:
                image = fits.open(image_path)[0].data
                if (np.shape(image)[0] != nx) or (np.shape(image)[1] != ny):
                    print(
                        f"image shape {np.shape(image)} is different from the coolest file  {nx}, {ny}"
                    )
            except:
                image = image_path
                print(
                    f"could not find image file {image_path}. Saving file name instead."
                )
            ra_at_xy_0 = -(
                list(lens_observation.pixels.field_of_view_x)[0] + pixel_size / 2.0
            )
            dec_at_xy_0 = (
                list(lens_observation.pixels.field_of_view_y)[0] + pixel_size / 2.0
            )
            transform_pix2angle = np.array([[-1, 0], [0, 1]]) * pixel_size

            kwargs_data = {
                "ra_at_xy_0": ra_at_xy_0,
                "dec_at_xy_0": dec_at_xy_0,
                "transform_pix2angle": transform_pix2angle,
                "image_data": image,
            }
            print("Data creation")

        # NOISE
        if lens_observation.noise is not None:
            if lens_observation.noise.type == "NoiseMap":
                creation_data = True
                noise_path = lens_observation.noise.noise_map.fits_file.path
                try:
                    noise = fits.open(noise_path)[0].data
                except:
                    noise = noise_path
                    print(
                        f"could not find noise file {noise_path}. Saving file name instead."
                    )
                noise_pixel_size = lens_observation.noise.noise_map.pixel_size
                noise_nx = lens_observation.noise.noise_map.num_pix_x
                noise_ny = lens_observation.noise.noise_map.num_pix_y
                if pixel_size != noise_pixel_size:
                    print(
                        f"noise pixel size {noise_pixel_size} is different from image pixel size {pixel_size}"
                    )
                if nx != noise_nx:
                    print(f"noise nx {noise_nx} is different from image nx {nx}")
                if ny != noise_ny:
                    print(f"noise ny {noise_ny} is different from image ny {ny}")
                kwargs_data["noise_map"] = noise
                print("Noise (in Data) creation")
            else:
                print(f"noise type {lens_observation.noise.type} is unknown")

    # PSF
    if lens_coolest.instrument is not None:
        lens_instrument = lens_coolest.instrument
        if lens_instrument.psf is not None:
            if lens_instrument.psf.type == "PixelatedPSF":
                creation_instrument = True
                psf_path = lens_instrument.psf.pixels.fits_file.path
                try:
                    psf = fits.open(psf_path)[0].data
                except:
                    psf = psf_path
                    print(
                        f"could not find PSF file {psf_path}. Saving file name instead."
                    )
                psf_pixel_size = lens_instrument.psf.pixels.pixel_size
                psf_nx = lens_instrument.psf.pixels.num_pix_x
                psf_ny = lens_instrument.psf.pixels.num_pix_y
                super_sampling_factor = 1
                if pixel_size != psf_pixel_size:
                    super_sampling_factor = int(pixel_size / psf_pixel_size)
                    print(
                        f"PSF pixel size {psf_pixel_size} is different from image pixel size {pixel_size}. "
                        f"Assuming super sampling factor of {super_sampling_factor}."
                    )

                kwargs_psf = {
                    "psf_type": "PIXEL",
                    "kernel_point_source": psf,
                    "point_source_supersampling_factor": super_sampling_factor,
                }
                print("PSF creation")
            else:
                print(f"PSF type {lens_instrument.psf.type} is unknown")

    # COSMO
    if lens_coolest.cosmology is not None:
        lens_cosmo = lens_coolest.cosmology
        if lens_cosmo.astropy_name == "FlatLambdaCDM":
            cosmo = FlatLambdaCDM(lens_cosmo.H0, lens_cosmo.Om0)
            creation_cosmo = True
            print("Cosmo class creation")
        else:
            print(f"Cosmology name {lens_cosmo.astropy_name} is unknown")

    # LIKELIHOODS not yet well supported by COOLEST
    # # LIKELIHOODS
    # if "likelihoods" not in exclude_keys:
    #     likelihoods = lens_coolest.likelihoods
    #     if likelihoods is not None:
    #         kwargs_likelihood = {}
    #         creation_kwargs_likelihood = True
    #         for like in likelihoods:
    #             if like == 'imaging_data':
    #                 kwargs_likelihood['image_likelihood'] = True
    #             else:
    #                 print(f"Likelihood {like} not yet implemented")
    #         print("kwargs_likelihood creation")

    # LENSING ENTITIES
    if lens_coolest.lensing_entities is not None:
        lensing_entities_list = lens_coolest.lensing_entities

        lens_model_list = []
        kwargs_lens = []
        kwargs_lens_up = []
        kwargs_lens_down = []
        kwargs_lens_init = []
        kwargs_lens_fixed = []
        kwargs_lens_sigma = []
        lens_light_model_list = []
        kwargs_lens_light = []
        kwargs_lens_light_up = []
        kwargs_lens_light_down = []
        kwargs_lens_light_init = []
        kwargs_lens_light_fixed = []
        kwargs_lens_light_sigma = []
        source_model_list = []
        kwargs_source = []
        kwargs_source_up = []
        kwargs_source_down = []
        kwargs_source_init = []
        kwargs_source_fixed = []
        kwargs_source_sigma = []
        ps_model_list = []
        kwargs_ps = []
        kwargs_ps_up = []
        kwargs_ps_down = []
        kwargs_ps_init = []
        kwargs_ps_fixed = []
        kwargs_ps_sigma = []

        creation_lens_source_light = True
        multi_plane = False
        creation_redshift_list = True

        min_redshift, max_redshift, redshift_list = create_redshift_info(
            lensing_entities_list
        )

        for lensing_entity in lensing_entities_list:
            if lensing_entity.type == "galaxy":
                galaxy = lensing_entity
                if galaxy.redshift > min_redshift:
                    # SOURCE OF LIGHT
                    light_list = galaxy.light_model
                    for light in light_list:
                        print("Source Light : ")
                        if light.type == "Sersic":
                            read.update_kwargs_sersic(
                                light,
                                source_model_list,
                                kwargs_source,
                                kwargs_source_init,
                                kwargs_source_up,
                                kwargs_source_down,
                                kwargs_source_fixed,
                                kwargs_source_sigma,
                                cleaning=True,
                            )
                        elif light.type == "Shapelets":
                            read.update_kwargs_shapelets(
                                light,
                                source_model_list,
                                kwargs_source,
                                kwargs_source_init,
                                kwargs_source_up,
                                kwargs_source_down,
                                kwargs_source_fixed,
                                kwargs_source_sigma,
                                cleaning=True,
                            )
                        elif light.type == "LensedPS":
                            read.update_kwargs_lensed_ps(
                                light,
                                ps_model_list,
                                kwargs_ps,
                                kwargs_ps_init,
                                kwargs_ps_up,
                                kwargs_ps_down,
                                kwargs_ps_fixed,
                                kwargs_ps_sigma,
                                cleaning=True,
                            )
                        else:
                            print(f"Light Type {light.type} not yet implemented.")

                if galaxy.redshift < max_redshift:
                    # LENSING GALAXY
                    if galaxy.redshift > min_redshift:
                        multi_plane = True
                        print("Multiplane lensing to consider.")
                    mass_list = galaxy.mass_model
                    for mass in mass_list:
                        print("Lens Mass : ")
                        if mass.type == "PEMD":
                            read.update_kwargs_pemd(
                                mass,
                                lens_model_list,
                                kwargs_lens,
                                kwargs_lens_init,
                                kwargs_lens_up,
                                kwargs_lens_down,
                                kwargs_lens_fixed,
                                kwargs_lens_sigma,
                                cleaning=True,
                            )
                        elif mass.type == "SIE":
                            read.update_kwargs_sie(
                                mass,
                                lens_model_list,
                                kwargs_lens,
                                kwargs_lens_init,
                                kwargs_lens_up,
                                kwargs_lens_down,
                                kwargs_lens_fixed,
                                kwargs_lens_sigma,
                                cleaning=True,
                            )
                        else:
                            print(f"Mass Type {mass.type} not yet implemented.")

                if galaxy.redshift == min_redshift:
                    # LENSING LIGHT GALAXY
                    light_list = galaxy.light_model
                    for light in light_list:
                        print("Lens Light : ")
                        if light.type == "Sersic":
                            read.update_kwargs_sersic(
                                light,
                                lens_light_model_list,
                                kwargs_lens_light,
                                kwargs_lens_light_init,
                                kwargs_lens_light_up,
                                kwargs_lens_light_down,
                                kwargs_lens_light_fixed,
                                kwargs_lens_light_sigma,
                                cleaning=True,
                            )
                        # elif light.type == 'LensedPS':
                        #     read.update_kwargs_lensed_ps(light, ps_model_list, kwargs_ps, kwargs_ps_init, kwargs_ps_up,
                        #                         kwargs_ps_down, kwargs_ps_fixed, kwargs_ps_sigma, cleaning=True)
                        else:
                            print(f"Light Type {light.type} not yet implemented.")

                # if (galaxy.redshift <= min_redshift) or (galaxy.redshift >= max_redshift):
                #     print(f'REDSHIFT {galaxy.redshift} is not in the range ] {min_red} , {max_red} [')

            elif lensing_entity.type == "MassField":
                mass_field_list = lensing_entity.mass_model
                for mass_field_idx in mass_field_list:
                    print("Shear : ")
                    if mass_field_idx.type == "ExternalShear":
                        read.update_kwargs_shear(
                            mass_field_idx,
                            lens_model_list,
                            kwargs_lens,
                            kwargs_lens_init,
                            kwargs_lens_up,
                            kwargs_lens_down,
                            kwargs_lens_fixed,
                            kwargs_lens_sigma,
                            cleaning=True,
                        )

                    else:
                        print(f"type of Shear {mass_field_idx.type} not implemented")

            else:
                print(f"lensing entity of type {lensing_entity.type} is unknown.")

    return_dict = {}
    if creation_lens_source_light is True:
        return_dict["kwargs_model"] = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "point_source_model_list": ps_model_list,
        }

        lens_params = [
            kwargs_lens_init,
            kwargs_lens_sigma,
            kwargs_lens_fixed,
            kwargs_lens_down,
            kwargs_lens_up,
        ]
        source_params = [
            kwargs_source_init,
            kwargs_source_sigma,
            kwargs_source_fixed,
            kwargs_source_down,
            kwargs_source_up,
        ]
        lens_light_params = [
            kwargs_lens_light_init,
            kwargs_lens_light_sigma,
            kwargs_lens_light_fixed,
            kwargs_lens_light_down,
            kwargs_lens_light_up,
        ]
        ps_params = [
            kwargs_ps_init,
            kwargs_ps_sigma,
            kwargs_ps_fixed,
            kwargs_ps_down,
            kwargs_ps_up,
        ]

        kwargs_params = {
            "lens_model": lens_params,
            "source_model": source_params,
            "lens_light_model": lens_light_params,
            "point_source_model": ps_params,
        }
        return_dict["kwargs_params"] = kwargs_params

        kwargs_result = {
            "kwargs_lens": kwargs_lens,
            "kwargs_source": kwargs_source,
            "kwargs_lens_light": kwargs_lens_light,
            "kwargs_ps": kwargs_ps,
        }
        return_dict["kwargs_result"] = kwargs_result
    if creation_redshift_list is True:
        return_dict["redshift_list"] = redshift_list
    # if creation_kwargs_likelihood is True:
    #     return_dict['kwargs_likelihood'] = kwargs_likelihood
    if creation_cosmo is True:
        return_dict["Cosmo"] = cosmo
    if creation_data is True:
        return_dict["kwargs_data"] = kwargs_data
    if creation_instrument is True:
        return_dict["kwargs_psf"] = kwargs_psf

    # time delay not implemented

    # I have never dealt with multiplane lensing :/ hope the redshift list is sufficient.
    # Need to implement more lens models/ source models.

    # GOOD TO KNOW : the COOLEST conventions have default bounds for some parameters
    # -> not updated with my default bounds !

    return return_dict


def update_coolest_from_lenstronomy(
    file_name, kwargs_result, kwargs_mcmc=None, ending="_update"
):
    """Function to update a json file already containing a model with the results of
    this model fitting.

    :param file_name: str, name (with path) of the json file to update
    :param kwargs_results: dict, lenstronomy kwargs_results {'kwargs_lens': [{..},{..}], 'kwargs_source': [{..}],...}
    :param kwargs_mcmc: dict, {'args_lens':args_lens,'args_source':args_source,'args_lens_light':args_lens_light,
                        'args_ps': args_ps}
                        with args_lens being a list, each element of the list being lens results of type
                        kwargs_results['kwargs_lens'] for a given MCMC point.
                        ex: {'args_lens': [[{'theta_E':0.71,...},{'gamma1':...}],
                                           [{'theta_E':0.709,...},{'gamma1':...}],
                                           [{'theta_E':0.711,...},{'gamma1':...}], ...],
                             'args_source': [[{'R_sersic':0.11,...}], [{'R_sersic':0.115,...}, ...]], ...}
                 if None, will not perform the Posterior update
    :param ending: str, ending of the name for saving the updated json file

     :return:   the new json file is saved with the updated kwargs.
    """

    decoder = JSONSerializer(file_name, indent=2)
    lens_coolest = decoder.load()
    available_profiles = [
        "LensedPS",
        "Sersic",
        "Shapelets",
        "PEMD",
        "SIE",
        "SIS",
        "ExternalShear",
    ]
    if lens_coolest.mode == "MAP":
        print(f"LENS COOLEST : {lens_coolest.mode}")
    else:
        print(f"LENS COOLEST IS NOT MAP, BUT IS {lens_coolest.mode}. CHANGING INTO MAP")
        lens_coolest.mode = "MAP"

    lensing_entities_list = lens_coolest.lensing_entities

    if lensing_entities_list is not None:
        lens_model_list = []
        kwargs_lens = []
        kwargs_lens_up = []
        kwargs_lens_down = []
        kwargs_lens_init = []
        kwargs_lens_fixed = []
        kwargs_lens_sigma = []
        lens_light_model_list = []
        kwargs_lens_light = []
        kwargs_lens_light_up = []
        kwargs_lens_light_down = []
        kwargs_lens_light_init = []
        kwargs_lens_light_fixed = []
        kwargs_lens_light_sigma = []
        source_model_list = []
        kwargs_source = []
        kwargs_source_up = []
        kwargs_source_down = []
        kwargs_source_init = []
        kwargs_source_fixed = []
        kwargs_source_sigma = []
        kwargs_ps = []
        kwargs_ps_up = []
        kwargs_ps_down = []
        kwargs_ps_init = []
        kwargs_ps_fixed = []
        kwargs_ps_sigma = []
        creation_lens_source_light = True

        idx_lens = 0
        idx_lens_light = 0
        idx_source = 0
        idx_ps = 0

        multi_plane = False
        creation_redshift_list = True

        min_redshift, max_redshift, redshift_list = create_redshift_info(
            lensing_entities_list
        )

        for lensing_entity in lensing_entities_list:
            if lensing_entity.type == "galaxy":
                galaxy = lensing_entity

                if galaxy.redshift > min_redshift:
                    # SOURCE OF LIGHT
                    light_list = galaxy.light_model
                    for light in light_list:
                        # ASSUME same list of models as in the json !!
                        if light.type == "LensedPS":
                            kwargs_ps = kwargs_result["kwargs_ps"][idx_ps]
                            kwargs_ps_mcmc = None
                        elif light.type in ["Sersic", "Shapelets"]:
                            kwargs_source = kwargs_result["kwargs_source"][idx_source]
                            kwargs_source_mcmc = None
                        else:
                            print(f"Light Type {light.type} not yet implemented.")

                        if (kwargs_mcmc is not None) & (
                            light.type in available_profiles
                        ):
                            if light.type == "LensedPS":
                                kwargs_ps_mcmc = [
                                    arg[idx_ps] for arg in kwargs_mcmc["args_ps"]
                                ]
                            elif light.type in ["Sersic", "Shapelets"]:
                                kwargs_source_mcmc = [
                                    arg[idx_source]
                                    for arg in kwargs_mcmc["args_source"]
                                ]

                        if light.type == "Sersic":
                            update.sersic_update(
                                light, kwargs_source, kwargs_source_mcmc
                            )
                            idx_source += 1
                        elif light.type == "Shapelets":
                            update.shapelets_update(
                                light, kwargs_source, kwargs_source_mcmc
                            )
                            idx_source += 1
                        elif light.type == "LensedPS":
                            update.lensed_point_source_update(
                                light, kwargs_ps, kwargs_ps_mcmc
                            )
                            idx_ps += 1
                        else:
                            pass

                if galaxy.redshift < max_redshift:
                    # LENSING GALAXY
                    if galaxy.redshift > min_redshift:
                        multi_plane = True
                        print("Multiplane lensing to consider.")
                    mass_list = galaxy.mass_model
                    for mass in mass_list:
                        kwargs_lens_mcmc = None
                        if (kwargs_mcmc is not None) & (
                            mass.type in available_profiles
                        ):
                            kwargs_lens_mcmc = [
                                arg[idx_lens] for arg in kwargs_mcmc["args_lens"]
                            ]

                        if mass.type == "PEMD":
                            kwargs_lens = kwargs_result["kwargs_lens"][idx_lens]
                            update.pemd_update(mass, kwargs_lens, kwargs_lens_mcmc)
                            idx_lens += 1
                        elif mass.type == "SIE":
                            kwargs_lens = kwargs_result["kwargs_lens"][idx_lens]
                            update.sie_update(mass, kwargs_lens, kwargs_lens_mcmc)
                            idx_lens += 1
                        else:
                            print(f"Mass Type {mass.type} not yet implemented.")

                if galaxy.redshift == min_redshift:
                    # LENSING LIGHT GALAXY
                    light_list = galaxy.light_model
                    for light in light_list:
                        if light.type in ["Sersic"]:
                            kwargs_lens_light = kwargs_result["kwargs_lens_light"][
                                idx_lens_light
                            ]
                            kwargs_lens_light_mcmc = None
                        else:
                            print(f"Light Type {light.type} not yet implemented.")

                        if (kwargs_mcmc is not None) & (
                            light.type in available_profiles
                        ):
                            if light.type in ["Sersic"]:
                                kwargs_lens_light_mcmc = [
                                    arg[idx_lens_light]
                                    for arg in kwargs_mcmc["args_lens_light"]
                                ]

                        if light.type == "Sersic":
                            update.sersic_update(
                                light, kwargs_lens_light, kwargs_lens_light_mcmc
                            )
                            idx_lens_light += 1
                        else:
                            pass
                #
                # if (galaxy.redshift <= min_redshift) and (galaxy.redshift >= max_redshift):
                #     print(f'REDSHIFT {galaxy.redshift} is not in the range ] {min_red} , {max_red} [')

            elif lensing_entity.type == "MassField":
                mass_field_list = lensing_entity.mass_model
                for mass_field_idx in mass_field_list:
                    kwargs_lens = kwargs_result["kwargs_lens"][idx_lens]
                    kwargs_lens_mcmc = None
                    if (kwargs_mcmc is not None) & (
                        mass_field_idx.type in available_profiles
                    ):
                        kwargs_lens_mcmc = [
                            arg[idx_lens] for arg in kwargs_mcmc["args_lens"]
                        ]

                    if mass_field_idx.type == "ExternalShear":
                        update.shear_update(
                            mass_field_idx, kwargs_lens, kwargs_lens_mcmc
                        )
                        idx_lens += 1
                    else:
                        print(f"type of Shear {mass_field_idx.type} not implemented")

            else:
                print(f"Lensing entity of type {lensing_entity.type} is unknown.")

    encoder = JSONSerializer(file_name + ending, obj=lens_coolest, indent=2)
    lens_coolest_encoded = encoder.dump_jsonpickle()

    return


def create_kwargs_mcmc_from_chain_list(
    chain_list,
    kwargs_model,
    kwargs_params,
    kwargs_data,
    kwargs_psf,
    kwargs_numerics,
    kwargs_constraints,
    image_likelihood_mask=None,
    idx_chain=-1,
    likelihood_threshold=None,
):
    """Function to construct kwargs_mcmc in the right format for the
    "update_coolest_from_lenstronomy" function.

    :param chain_list: list, output of FittingSequence.fitting_sequence()
    :param kwargs_model: the usual lenstronomy kwargs
    :param kwargs_params: the usual lenstronomy kwargs
    :param kwargs_data: the usual lenstronomy kwargs
    :param kwargs_psf: the usual lenstronomy kwargs
    :param kwargs_numerics: the usual lenstronomy kwargs
    :param kwargs_constraints: the usual lenstronomy kwargs
    :param image_likelihood_mask: the usual lenstronomy kwargs
    :param idx_chain: int, index of the MCMC chain in the chain_list, default is the
        last one. Can be useful if several PSO and MCMC are perfomed in the fitting
        sequence.
    :param likelihood_threshold: float, likelihood limit (negative) underwhich the MCMC
        point is not considered. Can be useful if a few chains are stucked in another
        (less good) minimum
    :return: kwargs_mcmc, list containing all the relevant MCMC points in a userfriendly
        format (with linear parameters etc)
    """
    par_buf = chain_list[idx_chain][1]
    dist_buf = chain_list[idx_chain][3]

    (
        kwargs_lens_init,
        kwargs_lens_sigma,
        kwargs_fixed_lens,
        kwargs_lower_lens,
        kwargs_upper_lens,
    ) = kwargs_params["lens_model"]
    (
        kwargs_source_init,
        kwargs_source_sigma,
        kwargs_fixed_source,
        kwargs_lower_source,
        kwargs_upper_source,
    ) = kwargs_params["source_model"]
    (
        kwargs_lens_light_init,
        kwargs_lens_light_sigma,
        kwargs_fixed_lens_light,
        kwargs_lower_lens_light,
        kwargs_upper_lens_light,
    ) = kwargs_params["lens_light_model"]
    (
        kwargs_ps_init,
        kwargs_ps_sigma,
        kwargs_fixed_ps,
        kwargs_lower_ps,
        kwargs_upper_ps,
    ) = kwargs_params["point_source_model"]

    param_class = Param(
        kwargs_model,
        kwargs_fixed_lens=kwargs_fixed_lens,
        kwargs_fixed_source=kwargs_fixed_source,
        kwargs_fixed_lens_light=kwargs_fixed_lens_light,
        kwargs_fixed_ps=kwargs_fixed_ps,
        kwargs_lower_lens=kwargs_lower_lens,
        kwargs_lower_source=kwargs_lower_source,
        kwargs_lower_lens_light=kwargs_lower_lens_light,
        kwargs_lower_ps=kwargs_lower_ps,
        kwargs_upper_lens=kwargs_upper_lens,
        kwargs_upper_source=kwargs_upper_source,
        kwargs_upper_lens_light=kwargs_upper_lens_light,
        kwargs_upper_ps=kwargs_upper_ps,
        kwargs_lens_init=kwargs_lens_init,
        **kwargs_constraints,
    )

    image_linear = class_util.create_image_model(
        kwargs_data,
        kwargs_psf,
        kwargs_numerics,
        kwargs_model,
        image_likelihood_mask=image_likelihood_mask,
    )
    args_lens = []
    args_source = []
    args_lens_light = []
    args_ps = []
    for w in range(len(dist_buf)):
        if likelihood_threshold is not None:
            if dist_buf[w] < likelihood_threshold:
                pass
        kwargs_return = param_class.args2kwargs(par_buf[w])
        image_linear.image_linear_solve(**kwargs_return)
        args_lens.append(kwargs_return["kwargs_lens"])
        args_source.append(kwargs_return["kwargs_source"])
        args_lens_light.append(kwargs_return["kwargs_lens_light"])
        args_ps.append(kwargs_return["kwargs_ps"])
    kwargs_mcmc_results = {
        "args_lens": args_lens,
        "args_source": args_source,
        "args_lens_light": args_lens_light,
        "args_ps": args_ps,
    }
    return kwargs_mcmc_results


def create_redshift_info(lensing_entities_list):
    """Side fuction to create the minimum, maximum and whole redshift list of galaxies
    in the COOLEST template Note that the redshifts helps knowing which galaxy is a
    lens, or a source, and if multiplane has to be considered.

    :param lensing_entities_list: coolest.template.classes.lensing_entity_list.LensingEntityList object
    :return: min_redshift, max_redshift, redshift_list ; minimum, maximum and full list of lensing entities redshifts
    """
    redshift_list = []
    for lensing_entity in lensing_entities_list:
        redshift_list.append(lensing_entity.redshift)
    min_redshift = np.min(redshift_list)
    max_redshift = np.max(redshift_list)
    return min_redshift, max_redshift, redshift_list



from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from astropy.cosmology import default_cosmology
from lenstronomy.Cosmo.lens_cosmo import LensCosmo

import copy

__all__ = ["ModelAPI"]


class ModelAPI(object):
    """This class manages the model choices.

    The role is to return instances of the lenstronomy LightModel, LensModel,
    PointSource modules according to the options chosen by the user. Currently, all
    other model choices are equivalent to the ones provided by LightModel, LensModel,
    PointSource. The current options of the class instance only describe a subset of
    possibilities.
    """

    def __init__(
        self,
        lens_model_list=None,
        z_lens=None,
        z_source=None,
        lens_redshift_list=None,
        source_light_model_list=None,
        lens_light_model_list=None,
        point_source_model_list=None,
        source_redshift_list=None,
        cosmo=None,
        z_source_convention=None,
        tabulated_deflection_angles=None,
        observed_convention_index=None,
    ):
        """# TODO: make inputs follow the kwargs_model of the class_creator instances of
        'kwargs_model', # i.e. multi-plane options, perhaps others

        :param lens_model_list: list of strings with lens model names
        :param z_lens: redshift of the deflector (only considered when operating in
            single plane mode). Is only needed for specific functions that require a
            cosmology.
        :param z_source: redshift of the source: Needed in multi_plane option only, not
            required for the core functionalities in the single plane mode. This will be
            the redshift of the source plane (if not further specified the
            'source_redshift_list') and the point source redshift (regardless of
            'source_redshift_list')
        :param lens_redshift_list: list of deflector redshift (corresponding to the lens
            model list), only applicable in multi_plane mode.
        :param source_light_model_list: list of strings with source light model names
            (lensed light profiles)
        :param lens_light_model_list: list of strings with lens light model names (not
            lensed light profiles)
        :param point_source_model_list: list of strings with point source model names
        :param source_redshift_list: list of redshifts of the source profiles (optional)
        :param cosmo: instance of the astropy cosmology class. If not specified, uses
            the default cosmology.
        :param z_source_convention: float, redshift of a source to define the reduced
            deflection angles of the lens models. If None, 'z_source' is used.
        :param tabulated_deflection_angles: a class that returns deflection angles given
            a set of (x, y) coordinates. Effectively a fixed lens model. See
            documentation in Profiles.numerical_alpha
        :param observed_convention_index: a list of indices that correspond to lens
            models where the center_x,center_y values correspond to the observed (lensed
            positions), not the physical positions in space
        """
        if lens_model_list is None:
            lens_model_list = []
        if source_light_model_list is None:
            source_light_model_list = []
        if lens_light_model_list is None:
            lens_light_model_list = []
        if point_source_model_list is None:
            point_source_model_list = []
        if cosmo is None:
            cosmo = default_cosmology.get()

        # making sure multi_plane mode is activated when either lens or source redshift lists are provided
        if lens_redshift_list is not None or source_redshift_list is not None:
            multi_plane = True
        else:
            multi_plane = False

        if z_source_convention is None:
            z_source_convention = z_source

        self._lens_model_class = LensModel(
            lens_model_list=lens_model_list,
            z_source=z_source,
            z_lens=z_lens,
            lens_redshift_list=lens_redshift_list,
            multi_plane=multi_plane,
            cosmo=cosmo,
            z_source_convention=z_source_convention,
            numerical_alpha_class=tabulated_deflection_angles,
            observed_convention_index=observed_convention_index,
        )
        self._source_model_class = LightModel(
            light_model_list=source_light_model_list,
            source_redshift_list=source_redshift_list,
        )
        self._lens_light_model_class = LightModel(
            light_model_list=lens_light_model_list
        )
        fixed_magnification = [False] * len(point_source_model_list)
        for i, ps_type in enumerate(point_source_model_list):
            if ps_type == "SOURCE_POSITION":
                fixed_magnification[i] = True
        self._point_source_model_class = PointSource(
            point_source_type_list=point_source_model_list,
            lensModel=self._lens_model_class,
            fixed_magnification_list=fixed_magnification,
        )
        self._cosmo = cosmo
        self._z_source_convention = z_source_convention
        self._lens_redshift_list = lens_redshift_list
        self._z_lens = z_lens

    @property
    def lens_model_class(self):
        """

        :return: instance of lenstronomy LensModel class
        """
        return self._lens_model_class

    @property
    def lens_light_model_class(self):
        """

        :return: instance of lenstronomy LightModel class describing the non-lensed light profiles
        """
        return self._lens_light_model_class

    @property
    def source_model_class(self):
        """

        :return: instance of lenstronomy LightModel class describing the source light profiles
        """
        return self._source_model_class

    @property
    def point_source_model_class(self):
        """

        :return: instance of lenstronomy PointSource class describing the point sources (lensed and unlensed)
        """
        return self._point_source_model_class

    def physical2lensing_conversion(self, kwargs_mass):
        """

        :param kwargs_mass: list of keyword arguments of all the lens models. Einstein radius 'theta_E' are replaced by
         'sigma_v', velocity dispersion in km/s, 'alpha_Rs' and 'Rs' of NFW profiles are replaced by 'M200' and
         'concentration'
        :return: kwargs_lens in reduced deflection angles compatible with the lensModel instance of this module
        """
        kwargs_lens = copy.deepcopy(kwargs_mass)
        for i in range(len(kwargs_mass)):
            kwargs_mass_i = kwargs_mass[i]
            if self._lens_redshift_list is None:
                z_lens = self._z_lens
            else:
                z_lens = self._lens_redshift_list[i]
            lens_cosmo = LensCosmo(z_lens, self._z_source_convention, cosmo=self._cosmo)

            if "sigma_v" in kwargs_mass_i:
                sigma_v = kwargs_mass_i["sigma_v"]
                theta_E = lens_cosmo.sis_sigma_v2theta_E(sigma_v)
                kwargs_lens[i]["theta_E"] = theta_E
                del kwargs_lens[i]["sigma_v"]
            elif "M200" in kwargs_mass_i:
                M200 = kwargs_mass_i["M200"]
                c = kwargs_mass_i["concentration"]
                Rs, alpha_RS = lens_cosmo.nfw_physical2angle(M200, c)
                kwargs_lens[i]["Rs"] = Rs
                kwargs_lens[i]["alpha_Rs"] = alpha_RS
                del kwargs_lens[i]["M200"]
                del kwargs_lens[i]["concentration"]
        return kwargs_lens



import numpy as np
from lenstronomy.LensModel.Profiles.sie import SIE
from lenstronomy.LensModel.Profiles.convergence import Convergence
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.Util import param_util

__all__ = ["CurvedArcTanDiff"]


class CurvedArcTanDiff(LensProfileBase):
    """Curved arc model with an additional non-zero tangential stretch differential in
    tangential direction component.

    Observables are:
    - curvature radius (basically bending relative to the center of the profile)
    - radial stretch (plus sign) thickness of arc with parity (more generalized than the power-law slope)
    - tangential stretch (plus sign). Infinity means at critical curve
    - direction of curvature
    - position of arc

    Requirements:
    - Should work with other perturbative models without breaking its meaning (say when adding additional shear terms)
    - Must best reflect the observables in lensing
    - minimal covariances between the parameters, intuitive parameterization.
    """

    param_names = [
        "tangential_stretch",
        "radial_stretch",
        "curvature",
        "dtan_dtan",
        "direction",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "tangential_stretch": -100,
        "radial_stretch": -5,
        "curvature": 0.000001,
        "dtan_dtan": -10,
        "direction": -np.pi,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "tangential_stretch": 100,
        "radial_stretch": 5,
        "curvature": 100,
        "dtan_dtan": 10,
        "direction": np.pi,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self._sie = SIE(NIE=True)
        self._mst = Convergence()
        super(CurvedArcTanDiff, self).__init__()

    @staticmethod
    def stretch2sie_mst(
        tangential_stretch,
        radial_stretch,
        curvature,
        dtan_dtan,
        direction,
        center_x,
        center_y,
    ):
        """

        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param dtan_dtan: d(tangential_stretch) / d(tangential direction) / tangential stretch
        :param direction: float, angle in radian
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return: parameters in terms of a spherical SIS + MST resulting in the same observables
        """
        center_x_sis, center_y_sis = center_deflector(
            curvature, direction, center_x, center_y
        )
        r_curvature = 1.0 / curvature
        lambda_mst = 1.0 / radial_stretch
        kappa_ext = 1 - lambda_mst
        theta_E = r_curvature * (1.0 - radial_stretch / tangential_stretch)
        # analytic relation (see Birrer 2021)
        dlambda_tan_dr = (
            tangential_stretch / r_curvature * (1 - tangential_stretch / radial_stretch)
        )

        # translate tangential eigenvalue gradient in lens ellipticity
        dtan_dtan_ = dtan_dtan * tangential_stretch
        epsilon = np.abs(dtan_dtan_ / dlambda_tan_dr)
        # bound epsilon by (-1, 1)
        epsilon = np.minimum(epsilon, 0.99999)
        q = np.sqrt((1 - epsilon) / (1 + epsilon))

        if dtan_dtan_ < 0:
            phi = direction - np.pi / 4
        else:
            phi = direction + np.pi / 4
        e1_sie, e2_sie = param_util.phi_q2_ellipticity(phi, q)

        # ellipticity adopted Einstein radius to match local tangential and radial stretch
        factor = np.sqrt(1 + q**2) / np.sqrt(2 * q)
        theta_E_sie = theta_E * factor
        return theta_E_sie, e1_sie, e2_sie, kappa_ext, center_x_sis, center_y_sis

    def function(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        dtan_dtan,
        direction,
        center_x,
        center_y,
    ):
        """
        ATTENTION: there may not be a global lensing potential!

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param dtan_dtan: d(tangential_stretch) / d(tangential direction) / tangential stretch
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        lambda_mst = 1.0 / radial_stretch
        (
            theta_E_sie,
            e1_sie,
            e2_sie,
            kappa_ext,
            center_x_sis,
            center_y_sis,
        ) = self.stretch2sie_mst(
            tangential_stretch,
            radial_stretch,
            curvature,
            dtan_dtan,
            direction,
            center_x,
            center_y,
        )
        f_sis = self._sie.function(
            x, y, theta_E_sie, e1_sie, e2_sie, center_x_sis, center_y_sis
        )  # - self._sis.function(center_x, center_y, theta_E, center_x_sis, center_y_sis)
        alpha_x, alpha_y = self._sie.derivatives(
            center_x, center_y, theta_E_sie, e1_sie, e2_sie, center_x_sis, center_y_sis
        )
        f_sis_0 = alpha_x * (x - center_x) + alpha_y * (y - center_y)
        f_mst = self._mst.function(x, y, kappa_ext, ra_0=center_x, dec_0=center_y)
        return lambda_mst * (f_sis - f_sis_0) + f_mst

    def derivatives(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        dtan_dtan,
        direction,
        center_x,
        center_y,
    ):
        """

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param dtan_dtan: d(tangential_stretch) / d(tangential direction) / tangential stretch
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        lambda_mst = 1.0 / radial_stretch
        (
            theta_E_sie,
            e1_sie,
            e2_sie,
            kappa_ext,
            center_x_sis,
            center_y_sis,
        ) = self.stretch2sie_mst(
            tangential_stretch,
            radial_stretch,
            curvature,
            dtan_dtan,
            direction,
            center_x,
            center_y,
        )
        f_x_sis, f_y_sis = self._sie.derivatives(
            x, y, theta_E_sie, e1_sie, e2_sie, center_x_sis, center_y_sis
        )
        f_x0, f_y0 = self._sie.derivatives(
            center_x, center_y, theta_E_sie, e1_sie, e2_sie, center_x_sis, center_y_sis
        )
        f_x_mst, f_y_mst = self._mst.derivatives(
            x, y, kappa_ext, ra_0=center_x, dec_0=center_y
        )
        f_x = lambda_mst * (f_x_sis - f_x0) + f_x_mst
        f_y = lambda_mst * (f_y_sis - f_y0) + f_y_mst
        return f_x, f_y

    def hessian(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        dtan_dtan,
        direction,
        center_x,
        center_y,
    ):
        """

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param dtan_dtan: d(tangential_stretch) / d(tangential direction) / tangential stretch
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        lambda_mst = 1.0 / radial_stretch
        (
            theta_E_sie,
            e1_sie,
            e2_sie,
            kappa_ext,
            center_x_sis,
            center_y_sis,
        ) = self.stretch2sie_mst(
            tangential_stretch,
            radial_stretch,
            curvature,
            dtan_dtan,
            direction,
            center_x,
            center_y,
        )
        f_xx_sis, f_xy_sis, f_yx_sis, f_yy_sis = self._sie.hessian(
            x, y, theta_E_sie, e1_sie, e2_sie, center_x_sis, center_y_sis
        )
        f_xx_mst, f_xy_mst, f_yx_mst, f_yy_mst = self._mst.hessian(
            x, y, kappa_ext, ra_0=center_x, dec_0=center_y
        )
        return (
            lambda_mst * f_xx_sis + f_xx_mst,
            lambda_mst * f_xy_sis + f_xy_mst,
            lambda_mst * f_yx_sis + f_yx_mst,
            lambda_mst * f_yy_sis + f_yy_mst,
        )


def center_deflector(curvature, direction, center_x, center_y):
    """

    :param curvature: 1/curvature radius
    :param direction: float, angle in radian
    :param center_x: center of source in image plane
    :param center_y: center of source in image plane
    :return: center_sis_x, center_sis_y
    """
    center_x_sis = center_x - np.cos(direction) / curvature
    center_y_sis = center_y - np.sin(direction) / curvature
    return center_x_sis, center_y_sis



__author__ = "ntessore"

import numpy as np
import lenstronomy.Util.util as util
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.LensModel.Profiles.spp import SPP
from scipy.special import hyp2f1

__all__ = ["EPL", "EPLMajorAxis", "EPLQPhi"]


class EPL(LensProfileBase):
    """Elliptical Power Law mass profile.

    .. math::
        \\kappa(x, y) = \\frac{3-\\gamma}{2} \\left(\\frac{\\theta_{E}}{\\sqrt{q x^2 + y^2/q}} \\right)^{\\gamma-1}

    with :math:`\\theta_{E}` is the (circularized) Einstein radius,
    :math:`\\gamma` is the negative power-law slope of the 3D mass distributions,
    :math:`q` is the minor/major axis ratio,
    and :math:`x` and :math:`y` are defined in a coordinate system aligned with the major and minor axis of the lens.

    In terms of eccentricities, this profile is defined as

    .. math::
        \\kappa(r) = \\frac{3-\\gamma}{2} \\left(\\frac{\\theta'_{E}}{r \\sqrt{1 - e*\\cos(2*\\phi)}} \\right)^{\\gamma-1}

    with :math:`\\epsilon` is the ellipticity defined as

    .. math::
        \\epsilon = \\frac{1-q^2}{1+q^2}

    And an Einstein radius :math:`\\theta'_{\\rm E}` related to the definition used is

    .. math::
        \\left(\\frac{\\theta'_{\\rm E}}{\\theta_{\\rm E}}\\right)^{2} = \\frac{2q}{1+q^2}.

    The mathematical form of the calculation is presented by Tessore & Metcalf (2015), https://arxiv.org/abs/1507.01819.
    The current implementation is using hyperbolic functions. The paper presents an iterative calculation scheme,
    converging in few iterations to high precision and accuracy.

    A (faster) implementation of the same model using numba is accessible as 'EPL_NUMBA' with the iterative calculation
    scheme. An alternative implementation of the same model using a fortran code FASTELL is implemented as 'PEMD'
    profile.
    """

    param_names = ["theta_E", "gamma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 1.5,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 2.5,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.epl_major_axis = EPLMajorAxis()
        self.spp = SPP()
        super(EPL, self).__init__()

    def param_conv(self, theta_E, gamma, e1, e2):
        """Converts parameters as defined in this class to the parameters used in the
        EPLMajorAxis() class.

        :param theta_E: Einstein radius as defined in the profile class
        :param gamma: negative power-law slope
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :return: b, t, q, phi_G
        """
        if self._static is True:
            return self._b_static, self._t_static, self._q_static, self._phi_G_static
        return self._param_conv(theta_E, gamma, e1, e2)

    @staticmethod
    def _param_conv(theta_E, gamma, e1, e2):
        """Convert parameters from :math:`R = \\sqrt{q x^2 + y^2/q}` to :math:`R =
        \\sqrt{q^2 x^2 + y^2}`

        :param gamma: power law slope
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :return: critical radius b, slope t, axis ratio q, orientation angle phi_G
        """
        t = gamma - 1
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        b = theta_E * np.sqrt(q)
        return b, t, q, phi_G

    def set_static(self, theta_E, gamma, e1, e2, center_x=0, center_y=0):
        """

        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: profile center
        :param center_y: profile center
        :return: self variables set
        """
        self._static = True
        (
            self._b_static,
            self._t_static,
            self._q_static,
            self._phi_G_static,
        ) = self._param_conv(theta_E, gamma, e1, e2)

    def set_dynamic(self):
        """

        :return:
        """
        self._static = False
        if hasattr(self, "_b_static"):
            del self._b_static
        if hasattr(self, "_t_static"):
            del self._t_static
        if hasattr(self, "_phi_G_static"):
            del self._phi_G_static
        if hasattr(self, "_q_static"):
            del self._q_static

    def function(self, x, y, theta_E, gamma, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: profile center
        :param center_y: profile center
        :return: lensing potential
        """
        b, t, q, phi_G = self.param_conv(theta_E, gamma, e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f_ = self.epl_major_axis.function(x__, y__, b, t, q)
        # rotate back
        return f_

    def derivatives(self, x, y, theta_E, gamma, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        b, t, q, phi_G = self.param_conv(theta_E, gamma, e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f__x, f__y = self.epl_major_axis.derivatives(x__, y__, b, t, q)
        # rotate back
        f_x, f_y = util.rotate(f__x, f__y, -phi_G)
        return f_x, f_y

    def hessian(self, x, y, theta_E, gamma, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: profile center
        :param center_y: profile center
        :return: f_xx, f_xy, f_yx, f_yy
        """

        b, t, q, phi_G = self.param_conv(theta_E, gamma, e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f__xx, f__xy, f__yx, f__yy = self.epl_major_axis.hessian(x__, y__, b, t, q)
        # rotate back
        kappa = 1.0 / 2 * (f__xx + f__yy)
        gamma1__ = 1.0 / 2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = np.cos(2 * phi_G) * gamma1__ - np.sin(2 * phi_G) * gamma2__
        gamma2 = +np.sin(2 * phi_G) * gamma1__ + np.cos(2 * phi_G) * gamma2__
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    def mass_3d_lens(self, r, theta_E, gamma, e1=None, e2=None):
        """Computes the spherical power-law mass enclosed (with SPP routine)

        :param r: radius within the mass is computed
        :param theta_E: Einstein radius
        :param gamma: power-law slope
        :param e1: eccentricity component (not used)
        :param e2: eccentricity component (not used)
        :return: mass enclosed a 3D radius r.
        """
        return self.spp.mass_3d_lens(r, theta_E, gamma)

    def density_lens(self, r, theta_E, gamma, e1=None, e2=None):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: radius within the mass is computed
        :param theta_E: Einstein radius
        :param gamma: power-law slope
        :param e1: eccentricity component (not used)
        :param e2: eccentricity component (not used)
        :return: mass enclosed a 3D radius r
        """
        return self.spp.density_lens(r, theta_E, gamma)


class EPLMajorAxis(LensProfileBase):
    """This class contains the function and the derivatives of the elliptical power law.

    .. math::
        \\kappa = (2-t)/2 * \\left[\\frac{b}{\\sqrt{q^2 x^2 + y^2}}\\right]^t

    where with :math:`t = \\gamma - 1` (from EPL class) being the projected power-law slope of the convergence profile,
    critical radius b, axis ratio q.

    Tessore & Metcalf (2015), https://arxiv.org/abs/1507.01819
    """

    param_names = ["b", "t", "q", "center_x", "center_y"]

    def __init__(self):
        super(EPLMajorAxis, self).__init__()

    def function(self, x, y, b, t, q):
        """Returns the lensing potential.

        :param x: x-coordinate in image plane relative to center (major axis)
        :param y: y-coordinate in image plane relative to center (minor axis)
        :param b: critical radius
        :param t: projected power-law slope
        :param q: axis ratio
        :return: lensing potential
        """
        # deflection from method
        alpha_x, alpha_y = self.derivatives(x, y, b, t, q)

        # deflection potential, eq. (15)
        psi = (x * alpha_x + y * alpha_y) / (2 - t)

        return psi

    def derivatives(self, x, y, b, t, q):
        """Returns the deflection angles.

        :param x: x-coordinate in image plane relative to center (major axis)
        :param y: y-coordinate in image plane relative to center (minor axis)
        :param b: critical radius
        :param t: projected power-law slope
        :param q: axis ratio
        :return: f_x, f_y
        """
        # elliptical radius, eq. (5)
        Z = np.empty(np.shape(x), dtype=complex)
        Z.real = q * x
        Z.imag = y
        R = np.abs(Z)
        R = np.maximum(R, 0.000000001)

        # angular dependency with extra factor of R, eq. (23)
        R_omega = Z * hyp2f1(1, t / 2, 2 - t / 2, -(1 - q) / (1 + q) * (Z / Z.conj()))

        # deflection, eq. (22)
        alpha = 2 / (1 + q) * (b / R) ** t * R_omega

        # return real and imaginary part
        alpha_real = np.nan_to_num(alpha.real, posinf=10**10, neginf=-(10**10))
        alpha_imag = np.nan_to_num(alpha.imag, posinf=10**10, neginf=-(10**10))

        return alpha_real, alpha_imag

    def hessian(self, x, y, b, t, q):
        """Hessian matrix of the lensing potential.

        :param x: x-coordinate in image plane relative to center (major axis)
        :param y: y-coordinate in image plane relative to center (minor axis)
        :param b: critical radius
        :param t: projected power-law slope
        :param q: axis ratio
        :return: f_xx, f_yy, f_xy
        """
        R = np.hypot(q * x, y)
        R = np.maximum(R, 0.00000001)
        r = np.hypot(x, y)

        cos, sin = x / r, y / r
        cos2, sin2 = cos * cos * 2 - 1, sin * cos * 2

        # convergence, eq. (2)
        kappa = (2 - t) / 2 * (b / R) ** t
        kappa = np.nan_to_num(kappa, posinf=10**10, neginf=-(10**10))

        # deflection via method
        alpha_x, alpha_y = self.derivatives(x, y, b, t, q)

        # shear, eq. (17), corrected version from arXiv/corrigendum
        gamma_1 = (1 - t) * (alpha_x * cos - alpha_y * sin) / r - kappa * cos2
        gamma_2 = (1 - t) * (alpha_y * cos + alpha_x * sin) / r - kappa * sin2
        gamma_1 = np.nan_to_num(gamma_1, posinf=10**10, neginf=-(10**10))
        gamma_2 = np.nan_to_num(gamma_2, posinf=10**10, neginf=-(10**10))

        # second derivatives from convergence and shear
        f_xx = kappa + gamma_1
        f_yy = kappa - gamma_1
        f_xy = gamma_2

        return f_xx, f_xy, f_xy, f_yy


class EPLQPhi(LensProfileBase):
    """Class to model a EPL sampling over q and phi instead of e1 and e2."""

    param_names = ["theta_E", "gamma", "q", "phi", "center_x", "center_y"]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 1.5,
        "q": 0,
        "phi": -np.pi,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 2.5,
        "q": 1,
        "phi": np.pi,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self._EPL = EPL()
        super(EPLQPhi, self).__init__()

    def function(self, x, y, theta_E, gamma, q, phi, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param q: axis ratio
        :param phi: position angle
        :param center_x: profile center
        :param center_y: profile center
        :return: lensing potential
        """
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        return self._EPL.function(x, y, theta_E, gamma, e1, e2, center_x, center_y)

    def derivatives(self, x, y, theta_E, gamma, q, phi, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param q: axis ratio
        :param phi: position angle
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        return self._EPL.derivatives(x, y, theta_E, gamma, e1, e2, center_x, center_y)

    def hessian(self, x, y, theta_E, gamma, q, phi, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param q: axis ratio
        :param phi: position angle
        :param center_x: profile center
        :param center_y: profile center
        :return: f_xx, f_xy, f_yx, f_yy
        """
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        return self._EPL.hessian(x, y, theta_E, gamma, e1, e2, center_x, center_y)

    def mass_3d_lens(self, r, theta_E, gamma, q=None, phi=None):
        """Computes the spherical power-law mass enclosed (with SPP routine).

        :param r: radius within the mass is computed
        :param theta_E: Einstein radius
        :param gamma: power-law slope
        :param q: axis ratio (not used)
        :param phi: position angle (not used)
        :return: mass enclosed a 3D radius r.
        """
        return self._EPL.mass_3d_lens(r, theta_E, gamma)

    def density_lens(self, r, theta_E, gamma, q=None, phi=None):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: radius within the mass is computed
        :param theta_E: Einstein radius
        :param gamma: power-law slope
        :param q: axis ratio (not used)
        :param phi: position angle (not used)
        :return: mass enclosed a 3D radius r
        """
        return self._EPL.density_lens(r, theta_E, gamma)
